C
C -----------------------------------------------------------------------------------------
C      
      TYPE fluidAtPartType
        REAL(8) vx,vy,vz  ! hitrost tekocine / fluid velocity
        REAL(8) wx,wy,wz  ! vrtincnost / vorticity
        REAL(8) T         ! temperature
        REAL(8) gradVx(3),gradVy(3),gradVz(3) ! velocity gradients
        REAL(8) dvxdt,dvydt,dvzdt ! velocity time derivatives
        REAL(8) nhhx,nhhy,nhhz ! nabla (\vec H \cdot \vec H)
      END TYPE
C
C -----------------------------------------------------------------------------------------
C      
      TYPE fluidType
        REAL(8), POINTER :: vx(:),vy(:),vz(:) ! hitrost tekocine, fluid velocity
        REAL(8), POINTER :: wx(:),wy(:),wz(:)  ! vrtincnost, fluid vorticity
        REAL(8), POINTER ::  T(:)  ! temperature
        REAL(8), POINTER :: pdens(:)  ! particle density
        REAL(8), POINTER :: gradVx(:,:) ! dvx/dx, dvx/dy, dvx/dz
        REAL(8), POINTER :: gradVy(:,:) ! dvy/dx, dvy/dy, dvy/dz
        REAL(8), POINTER :: gradVz(:,:) ! dvz/dx, dvz/dy, dvz/dz
        REAL(8), POINTER :: dvxdt(:),dvydt(:),dvzdt(:) ! dvx/dt, dvy/dt, dvz/dt
      END TYPE
C
C -----------------------------------------------------------------------------------------
C      
      TYPE particleType
        INTEGER :: id       ! identifikacijska stevilka delca, particle ID number
        REAL(8) :: x,y,z    ! lokacija, particle location
        REAL(8) :: vx,vy,vz ! hitrost delca, particle velocity
        LOGICAL :: active   ! ali je v toku ali ga je odneslo ven ali se je posedel, TRUE=in the domain, FALSE=out of the domain or stuck at the wall
        LOGICAL :: still    ! ali je na miru (recimo na zici), vendar je se vedno noter, TRUE=in the domain, but stuck at the electrical wire
        INTEGER :: cell     ! celica v kateri je, mesh cell number, where particle is located
        REAL(8) :: St       ! Stokes number
        INTEGER :: StClass       ! Stokes number class
        REAL(8) :: dp       ! premer delca, particle diameter
        REAL(8) :: vsx,vsy,vsz ! settling velocity delca, particle settling velocity
        REAL(8) :: Rep      ! Particle Reynolds number = dp*f_u0/f_visk
        INTEGER :: ppno     ! Particle processor number, which processor handles the particle
        INTEGER :: iPeri    ! stevilka periode v kateri je (ce je periodicno), for periodic flow, which period is particle in
      END TYPE
C
C -----------------------------------------------------------------------------------------
C      
      TYPE CommonParticleDataType
        INTEGER :: maxNp  ! maksimalno stevilo delcev, maximal number of particles
        REAL(8) :: p_rho  ! gostota delcev [kg/m^3], particle density
        REAL(8) :: f_L    ! karakteristicna dolzina za tekocino [m], characteristic particle lenght scale
        REAL(8) :: f_visk ! viskoznost tekocine [m/s^2], fluid viscosity
        REAL(8) :: f_Re   ! Reynoldsovo stevilo za tekocino, fluid Reynolds number
        REAL(8) :: f_rho  ! gostota tekocine [kg/m^3], fluid denstiy
        REAL(8) :: f_u0   ! karakteristicna hitrost [m/s], characteristic fluid velocity
        REAL(8) :: R      ! = f_rho / ( p_rho + 0.5*f_rho)
        REAL(8) :: A      ! = p_rho / ( p_rho + 0.5*f_rho) = 1 / ( 1 + 0.5*R*(1+f_rho/2/p_rho))
                          ! A je izracunan tako da je = 0, ce postavim R=0 (da zanemarim added mass in pressure gradient clena) 
        REAL(8) :: g0     ! gravitacijski pospesek [m/s^2], gravitational acceleration
        REAL(8) :: gx,gy,gz ! smer gravitacije \vec g / |\vec g|, direction of gravitational acceleration
        INTEGER :: hepc   ! heavy particle correction
        INTEGER :: npartdesc ! stevilo navodil za vnos delcev, number of PART statements in input file
        CHARACTER(255), POINTER :: partdesc(:) ! navodila za vnos delcev, PART statements from input file
        INTEGER :: cnp    ! current number of particles
        INTEGER :: lid    ! last particle ID
        INTEGER :: simt  ! tip simulacije, simulation type
        INTEGER :: swg   ! search whole grid for particle
        
        REAL(8) :: maxSt   ! max Stokes izmed vseh delcev na zacetku
        REAL(8) :: minSt   ! min Stokes

        REAL(8) :: f_hi, p_hi ! magnetic susceptibility of fluid, particle
        REAL(8) :: H0       ! magnetic field strength [A/m]
        REAL(8) :: mfac     ! magnetic factor before nabla(H*H)
      END TYPE
C
C -----------------------------------------------------------------------------------------
C      
      TYPE meshType
        CHARACTER*255 ::  FileName
        CHARACTER*50  ::  Version
        CHARACTER*255 ::  Path
        CHARACTER*255 ::  Title        
        CHARACTER*255 ::  FullName            ! path & file name
        REAL(8) :: id                         ! unique ID
        INTEGER :: nbelem,nicell,nnodes,nbnodes,npob,npoc,npx !,nx,ny,simx,simy,nofc,
        INTEGER :: nofw
        INTEGER :: nside,npof  ! stevilo stranic = 6, stevilo nodeov na nezvezni stranici =4
        INTEGER :: nq      ! stevilo nezveznih nodeov (vseh skupaj)
        INTEGER :: npofc   !  stevilo nezveznih nodeov v eni celici = nside*npof=24
        INTEGER :: nicnpoc ! nicell * npoc 27
        INTEGER :: nsp     !=mesh%npoc+mesh%npofc   ! 27+24=51
        INTEGER :: nicnsp  !=nicell * nsp
        INTEGER :: nnmnb   ! = nnodes - nbnodes
        
        INTEGER, POINTER :: ibc(:,:)  ! nbelem,npob
        INTEGER, POINTER :: idc(:,:)  ! nicell,npoc
        INTEGER, POINTER :: ibf(:,:,:)! nicell,nside,npof
        INTEGER, POINTER :: ibcf(:,:) ! nbelem,npof        
        REAL(8), POINTER :: x(:,:)    ! nnodes,npx
        REAL(8), POINTER :: xq(:,:)   ! nq,npx  
        REAL(8), POINTER :: vol(:)   ! volumen celice      

        INTEGER, POINTER :: ukode(:) ! doloci ali je vozlice znano ali neznano
        INTEGER, POINTER :: qkode(:) ! na katero po vrsti je v vektrju

C       makro vorticity transport equation        
        INTEGER, POINTER :: wkode(:,:) ! doloci ali je vozlice znano ali neznano
        INTEGER, POINTER :: wqkode(:,:) ! na katero po vrsti je v vektrju
        INTEGER, POINTER :: nunk(:),nb(:) ! stevilo neznank in znanih vrednosti

C       makro energy transport equation        
        INTEGER, POINTER :: Tkode(:) ! doloci ali je vozlice znano ali neznano
        INTEGER, POINTER :: Tqkode(:) ! na katero po vrsti je v vektrju
        INTEGER Tnunk,Tnb ! stevilo neznank in znanih vrednosti

        
        INTEGER, POINTER :: flpm(:,:) ! flukc plus minus, katera G integral je s plusom kateri z munisim


        INTEGER, POINTER :: gbn(:)    ! boundary node list (nbnodes)
        
        INTEGER, POINTER :: lbn(:)    ! inverse boundary node list (nnodes), 0=obmocno vozlisce

        REAL(8), POINTER :: nx(:),ny(:),nz(:) ! noramala na tocko na robu (nbnodes)


        INTEGER, POINTER :: bewn(:)   ! boundary element wall number (nbelem)
        INTEGER, POINTER :: iwn(:)    ! boundary node wall number (nbnodes)

        INTEGER, POINTER  :: neil(:,:) ! list of cell neighbours

c        INTEGER, POINTER :: iwc(:,:)        
         INTEGER, POINTER :: kode(:,:)                
         INTEGER, POINTER :: kobc(:) 
c        REAL(8), POINTER :: kobcKsi(:),kobcEta(:)                    
c        INTEGER, POINTER :: icn(:)        
c        INTEGER, POINTER :: isw(:)        
c        REAL(8), POINTER :: intsize(:)         ! makro mesh interpolation, total size of cells node belongs to
       
        INTEGER, POINTER :: dnl(:)    ! domain node list (nnodes-nbnodes), 
        INTEGER, POINTER :: idnl(:)   ! inverse domain node list (nnodes), 0=robno vozlisce

        INTEGER, POINTER :: mfdom(:) ! nicell - kjer magnetic force dominates

      END TYPE meshType 
C
C -----------------------------------------------------------------------------------------
C
      TYPE ForceStatisticsType
        REAL(8), POINTER :: gb(:)
        REAL(8), POINTER :: drag(:)
        REAL(8), POINTER :: adpc(:) !added mass + pressure correction
        REAL(8), POINTER :: mag(:)
        REAL(8), POINTER :: vabs(:)
        REAL(8), POINTER :: aabs(:)
        INTEGER, POINTER :: n(:) ! number of data
      END TYPE

C
C -----------------------------------------------------------------------------------------
C            
      TYPE InputType      
        CHARACTER*100 IDname,IDversion,IDdate
        CHARACTER*39 StartTime,EndTime      
        CHARACTER*255 mdir,lgeo,lgeo_withpath
        CHARACTER*255 redf  ! results data file
        REAL(8) tstep ! naj bi bil desetina St stevila, time step
        INTEGER nsteps ! stevilo casovnih korakov , number of time steps
        INTEGER outst ! na koliko korakov naredimo izpis rezultatov, write resutls every OUTST's time step
        INTEGER rove  ! R overwrite - R=0, A=1, upostevamo samo g in drag., ROVE 1 = take into account only g and drag
        INTEGER cfxo  ! ali zapisem datoteko, ki omogoca animacijo delcev v cfx (prek dat2trk), CFX output
        INTEGER lvcs  ! List of visited cells, stevilo stokesovih razredov
        INTEGER imag  ! ali imamo magnetno silo, Do we take into account the magnetic force
        INTEGER mmodel ! katero magnetno polje izberemo, model for magnetic field
        REAL(8) xpi    ! za cell flow, koliko krat pi je v sinusu, parameter for cell flow
        INTEGER pdens  ! izpise particle density field, measure particle density field
        INTEGER ploc   ! izpise lokacije delcev po timestepih v tecplot in bin format, write tecplot output file
        INTEGER mfpid ! monitor forces particle id, monitor forces on particle

        INTEGER iwire, nwire ! stevilo zic, number of electrical wires
        REAL(8), POINTER :: wireX(:),wireZ(:),wireSign(:)
        INTEGER nPeri ! stevilo period, number of periods in periodic flows
        REAL(8) Xlength ! dolzina kanala v smeri x, channel lenght in X direction
        REAL(8) wr ! wire radius
      END TYPE       
C
C -----------------------------------------------------------------------------------------
C            
      TYPE GaussType
         INTEGER :: kmBs,kmBr,kmDs,kmDr    ! gre od 1 do 8=48)
         REAL(8), POINTER :: gi(:),ome(:)
         INTEGER, POINTER :: ng1(:),ng2(:)
      END TYPE   
C
C -----------------------------------------------------------------------------------------
C            
      TYPE IOtype     
        INTEGER l,inp,errwarn,mesh,redf,neil,trf,pl,ple,trf1,lvc,trf2,ibin,plr,ipdens,imf
        INTEGER mfdom,fs,md,ns
        CHARACTER*60 l_name,inp_name,errwarn_name,neil_name,trf_name,pl_name,ple_name,trf1_name,lvc_name,trf2_name
        CHARACTER*60 ibin_name,plr_name,ipdens_name,imf_name,mfdom_name,fs_name,md_name,ns_name
      END TYPE   

C
C -----------------------------------------------------------------------------------------
C
      INCLUDE 'mpif.h'
      TYPE penv
        INTEGER          :: comm        ! communicator number
        INTEGER          :: nproc       ! number of processes
        INTEGER          :: mpr         ! my process number (zero based)
        INTEGER          :: myproc      ! my process number (one based)
      END TYPE
C
