C -----------------------------------------------------------------------------
      SUBROUTINE ReadInputFile(mesh,io,inp,cpd)
C
C     $: reads bem.inp
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'
      TYPE(meshType) :: mesh
      TYPE(InputType) inp
      TYPE(IOtype) io
      TYPE(CommonParticleDataType) :: cpd  ! common particle data
      
      INTEGER lun,i
      CHARACTER KeyWord*4,OneLine*255,dummy*64,tmp*255

      lun=io%inp
      OPEN (lun,FILE=io%inp_name,ERR=10,STATUS='OLD',SHARED)

C
C***    Set up default settings:
c
      cpd%g0=9.81D0
      cpd%gx=0.0D0
      cpd%gy=0.0D0
      cpd%gz=-1.0D0
      cpd%npartdesc=0
      cpd%hepc=0  ! do not use heavy particle correction
      inp%tstep=0.001D0
      inp%nsteps=10
      inp%rove=0
      cpd%simt=1
      cpd%swg=0
      inp%cfxo=0 ! export results in CFX format (does not work anymore)
      inp%lvcs=1 ! export list of visited cells
      inp%imag=0
      inp%xpi=2.0D0
      inp%pdens=1 ! export tecplot particle density
      inp%ploc=1  ! export tecplot particle location
      inp%mfpid=1 ! monitor forces - prvi delec
      inp%iwire=0 ! ni noter zic

      inp%nPeri=1 ! stevilo period za kanal v smeri X
      inp%Xlength=1.0D0

      CALL rOneTL(lun,OneLine)
      DO WHILE (OneLine(1:3).NE.'EOF')
        READ (OneLine,*) KeyWord
C
C***    GEO AND BC FILE LOCATION :
c
        IF (KeyWord.EQ.'MDIR') THEN
          READ(OneLine,'(A5,A)') dummy,mesh%path

        ELSE IF (KeyWord.EQ.'LGEO') THEN
          READ(OneLine,'(A5,A)') dummy,mesh%filename

        ELSE IF (KeyWord.EQ.'REDF') THEN
          READ(OneLine,'(A5,A)') dummy,inp%redf

        ELSE IF (KeyWord.EQ.'MNOP') THEN
          READ(OneLine,*) dummy,cpd%maxNp

        ELSE IF (KeyWord.EQ.'PRHO') THEN
          READ(OneLine,*) dummy,cpd%p_rho

        ELSE IF (KeyWord.EQ.'HEPC') THEN
          READ(OneLine,*) dummy,cpd%hepc

        ELSE IF (KeyWord.EQ.'SIMT') THEN
          READ(OneLine,*) dummy,cpd%simt

        ELSE IF (KeyWord.EQ.'CELL') THEN
          READ(OneLine,*) dummy,inp%xpi  ! pazi, SIMT mora biti 2

        ELSE IF (KeyWord.EQ.'ROVE') THEN
          READ(OneLine,*) dummy,inp%rove

        ELSE IF (KeyWord.EQ.'MONF') THEN
          READ(OneLine,*) dummy,inp%mfpid

        ELSE IF (KeyWord.EQ.'LVCS') THEN
          READ(OneLine,*) dummy,inp%lvcs

        ELSE IF (KeyWord.EQ.'PDEN') THEN
          READ(OneLine,*) dummy,inp%pdens

        ELSE IF (KeyWord.EQ.'PLOC') THEN
          READ(OneLine,*) dummy,inp%ploc

        ELSE IF (KeyWord.EQ.'CFXO') THEN
          READ(OneLine,*) dummy,inp%cfxo

        ELSE IF (KeyWord.EQ.'WIRE') THEN
          READ(OneLine,*) dummy,inp%iwire,inp%nwire
          IF (inp%iwire.GT.0) THEN
            ALLOCATE (inp%wireX(inp%nwire))
            ALLOCATE (inp%wireZ(inp%nwire))
            ALLOCATE (inp%wireSign(inp%nwire))
            DO i=1,inp%nwire
              CALL rOneTL(lun,OneLine)
              READ(OneLine,*) inp%wireX(i),inp%wireZ(i),inp%wireSign(i)
            END DO
          END IF
        ELSE IF (KeyWord.EQ.'TSTE') THEN
          READ(OneLine,*) dummy,inp%tstep,inp%nsteps,inp%outst

        ELSE IF (KeyWord.EQ.'PERI') THEN
          READ(OneLine,*) dummy,inp%nPeri,inp%Xlength

        ELSE IF (KeyWord.EQ.'GRAV') THEN
          READ(OneLine,*) dummy,cpd%g0,cpd%gx,cpd%gy,cpd%gz

        ELSE IF (KeyWord.EQ.'MAGN') THEN
          READ(OneLine,*) dummy,inp%imag,inp%mmodel,cpd%f_hi,cpd%p_hi,cpd%H0

        ELSE IF (KeyWord.EQ.'FLUL') THEN
          READ(OneLine,*) dummy,cpd%f_Re,cpd%f_rho,cpd%f_visk,cpd%f_L
          cpd%f_u0=cpd%f_Re*cpd%f_visk/cpd%f_L

        ELSE IF (KeyWord.EQ.'FLUU') THEN
          READ(OneLine,*) dummy,cpd%f_Re,cpd%f_rho,cpd%f_visk,cpd%f_u0
          cpd%f_L=cpd%f_Re*cpd%f_visk/cpd%f_u0

        ELSE IF (KeyWord.EQ.'PART') THEN
          cpd%npartdesc=cpd%npartdesc+1
          READ(OneLine,'(A5,A)') dummy,cpd%partdesc(cpd%npartdesc)
          
        END IF
        CALL rOneTL(lun,OneLine)
      END DO

      CLOSE (lun)
      mesh%fullname=Trim(mesh%path)//Trim(mesh%filename)  

      inp%mdir=mesh%path
      inp%lgeo=mesh%filename
      inp%lgeo_withpath=mesh%fullname
      
c     common particle data, tu daj if za R overwrite, ce zelis zanemariti clene
      IF (inp%rove.EQ.1) THEN
        cpd%R=0.0D0
      ELSE
        cpd%R = cpd%f_rho / ( cpd%p_rho + 0.5D0*cpd%f_rho)
      END IF
      cpd%A = 1.0D0 / ( 1.0D0 + 0.5D0*cpd%R*(1.0D0+cpd%f_rho/2.0D0/cpd%p_rho))
    
      RETURN
      
10    continue ! error when opening input file
      WRITE (tmp,'(A,A)') "could not open : ",io%inp_name
      CALL WarnErr(io,inp,2,"ReadInputFile",trim(tmp),0)
      
      END


C -----------------------------------------------------------------------------
      SUBROUTINE Count_ReadInputFile(mesh,io,inp,cpd)
C
C     $: reads bem.inp
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'
      TYPE(meshType) :: mesh
      TYPE(InputType) inp
      TYPE(IOtype) io
      TYPE(CommonParticleDataType) :: cpd  ! common particle data

      INTEGER lun
      CHARACTER KeyWord*4,OneLine*255,tmp*255

      cpd%npartdesc=0
      lun=io%inp
      OPEN (lun,FILE=io%inp_name,ERR=10,STATUS='OLD',SHARED)

      CALL rOneTL(lun,OneLine)
      DO WHILE (OneLine(1:3).NE.'EOF')
        READ (OneLine,*) KeyWord
        IF (KeyWord.EQ.'PART') THEN
          cpd%npartdesc=cpd%npartdesc+1
        END IF
        CALL rOneTL(lun,OneLine)
      END DO

      CLOSE (lun)

      RETURN

10    continue ! error when opening input file
      WRITE (tmp,'(A,A)') "could not open : ",io%inp_name
      CALL WarnErr(io,inp,2,"ReadInputFile",trim(tmp),0)

      END

C -----------------------------------------------------------------------------
      SUBROUTINE Outlvc(mesh,io,inp,lvc)
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'
      TYPE(meshType) :: mesh
      TYPE(IOtype)    :: io
      TYPE(inputtype) :: inp     

      INTEGER lvc(mesh%nicell,inp%lvcs)
      REAL(8), ALLOCATABLE ::  nodelvc(:,:)
      INTEGER i,j,k,nic
      CHARACTER*255 vrstica      
      
      ALLOCATE (nodelvc(mesh%nnodes,inp%lvcs+1))
      nodelvc=0.0D0

      DO i=1,mesh%nicell
        DO k=1,inp%lvcs
          IF (lvc(i,k).GT.0) THEN
            DO j=1,mesh%npoc
              nodelvc(mesh%idc(i,j),k)=1.0D0
            END DO           
          END IF
        END DO
      END DO

c     skupen LVC, za vse delce ne glede na St stevilo      
      DO i=1,mesh%nnodes
        DO j=1,inp%lvcs
          IF (nodelvc(i,j).GT.0.0D0) nodelvc(i,inp%lvcs+1)=1.0D0
        END DO
      END DO
            

      nic=8*mesh%nicell
      OPEN (io%lvc,FILE=io%lvc_name,STATUS='UNKNOWN')

c     naslov, variables      
      vrstica=''
      DO i=1,inp%lvcs
        WRITE(vrstica,'(A,A,I2,A)') trim(vrstica),',"StClass',i,'"'
      END DO
      
      WRITE (io%lvc,'(A,A)') 'VARIABLES = "X", "Y", "Z","node","lvc"',vrstica
      
      WRITE (io%lvc,'(A,I8,A,I8)') 
     &      'ZONE T="visited cells",F=FEPOINT, ET=BRICK,N= ', mesh%NNODES, ',E=', nic
     
C       Output results        
        DO  i=1,mesh%nnodes
          WRITE (vrstica,'(3G18.9,I8,<inp%lvcs+1>F7.3)') 
     &        mesh%x(i,1),mesh%x(i,2),mesh%x(i,3),i,nodelvc(i,inp%lvcs+1),(nodelvc(i,j),j=1,inp%lvcs)
     
          CALL sqblnk(io%lvc,vrstica)           
        END DO
        CALL TECwIDC(io%lvc,mesh)
      
      CLOSE(io%lvc)
      
      DEALLOCATE (nodelvc)
      
      END
      
C -----------------------------------------------------------------------------
      SUBROUTINE OutMfdom(mesh,io,mfdom)
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'
      TYPE(meshType) :: mesh
      TYPE(IOtype)    :: io

      INTEGER mfdom(mesh%nicell)
      INTEGER, ALLOCATABLE ::  nodemfdom(:)
      INTEGER i,j,nic
      CHARACTER*255 vrstica

      ALLOCATE (nodemfdom(mesh%nnodes))
      nodemfdom=0.0D0

      DO i=1,mesh%nicell
        IF (mfdom(i).GT.0) THEN
          DO j=1,mesh%npoc
            nodemfdom(mesh%idc(i,j))=1.0D0
          END DO
        END IF
      END DO

      nic=8*mesh%nicell
      OPEN (io%mfdom,FILE=io%mfdom_name,STATUS='UNKNOWN')

c     naslov, variables
      WRITE (io%mfdom,'(A)') 'VARIABLES = "X", "Y", "Z","node","mfdom"'

      WRITE (io%mfdom,'(A,I8,A,I8)')
     &      'ZONE T="mfdom",F=FEPOINT, ET=BRICK,N= ', mesh%NNODES, ',E=', nic

C       Output results
      DO  i=1,mesh%nnodes
        WRITE (vrstica,'(3G18.9,I8,I5)') mesh%x(i,1),mesh%x(i,2),mesh%x(i,3),i,nodemfdom(i)
        CALL sqblnk(io%mfdom,vrstica)
      END DO
        CALL TECwIDC(io%mfdom,mesh)

      CLOSE(io%mfdom)

      DEALLOCATE (nodemfdom)

      END


C -----------------------------------------------------------------------------
      SUBROUTINE OutFstat(mesh,io,fstat)
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'
      TYPE(meshType) :: mesh
      TYPE(IOtype)    :: io

      TYPE(ForceStatisticsType) :: fstat
      REAL(8), ALLOCATABLE ::  nodegb(:)
      REAL(8), ALLOCATABLE ::  nodedrag(:)
      REAL(8), ALLOCATABLE ::  nodeadpc(:)
      REAL(8), ALLOCATABLE ::  nodemag(:)
      REAL(8), ALLOCATABLE ::  nodevabs(:)
      REAL(8), ALLOCATABLE ::  nodeaabs(:)
      REAL(8), ALLOCATABLE ::  noden(:)
      REAL(8) vsota
      INTEGER, ALLOCATABLE :: ndat(:)
      INTEGER i,j,nic
      CHARACTER*255 vrstica

      ALLOCATE (nodegb(mesh%nnodes))
      ALLOCATE (nodedrag(mesh%nnodes))
      ALLOCATE (nodeadpc(mesh%nnodes))
      ALLOCATE (nodemag(mesh%nnodes))
      ALLOCATE (nodevabs(mesh%nnodes))
      ALLOCATE (nodeaabs(mesh%nnodes))
      ALLOCATE (noden(mesh%nnodes))
      nodegb=0.0D0
      nodedrag=0.0D0
      nodeadpc=0.0D0
      nodemag=0.0D0
      nodevabs=0.0D0
      nodeaabs=0.0D0
      noden=0.0D0

      ALLOCATE (ndat(mesh%nnodes))
      ndat=0

      DO i=1,mesh%nicell
        vsota=fstat%gb(i)+fstat%drag(i)+fstat%adpc(i)+fstat%mag(i)
        IF (vsota.GT.0.0D0) THEN
          DO j=1,mesh%npoc
            ndat(mesh%idc(i,j))=ndat(mesh%idc(i,j))+1
            nodegb(mesh%idc(i,j))=nodegb(mesh%idc(i,j))+fstat%gb(i)
            nodedrag(mesh%idc(i,j))=nodedrag(mesh%idc(i,j))+fstat%drag(i)
            nodeadpc(mesh%idc(i,j))=nodeadpc(mesh%idc(i,j))+fstat%adpc(i)
            nodemag(mesh%idc(i,j))=nodemag(mesh%idc(i,j))+fstat%mag(i)
            nodevabs(mesh%idc(i,j))=nodevabs(mesh%idc(i,j))+fstat%vabs(i)
            nodeaabs(mesh%idc(i,j))=nodeaabs(mesh%idc(i,j))+fstat%aabs(i)
            noden(mesh%idc(i,j))=noden(mesh%idc(i,j))+DBLE(fstat%n(i))
          END DO
        END IF
      END DO

      DO i=1,mesh%nnodes
c       print *,ndat(i),nodegb(i)+nodedrag(i)+nodeadpc(i)+nodemag(i)
        IF (ndat(i).GT.0) THEN
          nodegb(i)=nodegb(i)/ndat(i)
          nodedrag(i)=nodedrag(i)/ndat(i)
          nodeadpc(i)=nodeadpc(i)/ndat(i)
          nodemag(i)=nodemag(i)/ndat(i)
          nodevabs(i)=nodevabs(i)/ndat(i)
          nodeaabs(i)=nodeaabs(i)/ndat(i)
          noden(i)=noden(i)/ndat(i)
        END IF
      END DO


      nic=8*mesh%nicell
      OPEN (io%fs,FILE=io%fs_name,STATUS='UNKNOWN')

c     naslov, variables
      WRITE (io%fs,'(A)') 'VARIABLES = "X", "Y", "Z","node","gb","drag","adpc","mag","vabs","aabs","n"'

      WRITE (io%fs,'(A,I8,A,I8)')
     &      'ZONE T="mfdom",F=FEPOINT, ET=BRICK,N= ', mesh%NNODES, ',E=', nic

C       Output results
      DO  i=1,mesh%nnodes
        WRITE (vrstica,'(3G18.9,I8,4F7.3,2G18.9,G18.5)')
     &    mesh%x(i,1),mesh%x(i,2),mesh%x(i,3),i,nodegb(i),nodedrag(i),nodeadpc(i),nodemag(i),nodevabs(i),nodeaabs(i),noden(i)
        CALL sqblnk(io%fs,vrstica)
      END DO
        CALL TECwIDC(io%fs,mesh)

      CLOSE(io%fs)

      END



C -----------------------------------------------------------------------------
      SUBROUTINE OutMdist(io,inp,mdist)
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'
      TYPE(IOtype)    :: io
      TYPE(inputtype) :: inp

      REAL(8) mdist(inp%nsteps),t
      INTEGER i
      CHARACTER*255 vrstica


      OPEN (io%md,FILE=io%md_name,STATUS='UNKNOWN')

c     naslov, variables
      WRITE (io%md,'(A)') 'VARIABLES = "time", "mdist"'
      WRITE (io%md,'(A)') 'ZONE T="mdist"'

C       Output results
      t=0.0D0
      DO i=1,inp%nsteps
        t=t+inp%tstep
        WRITE (vrstica,'(2G18.9)') t,mdist(i)
        CALL sqblnk(io%md,vrstica)
      END DO

      CLOSE(io%md)

      END

C -----------------------------------------------------------------------------
      SUBROUTINE OutNstill(io,inp,nstill,cpd)
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'
      TYPE(IOtype)    :: io
      TYPE(inputtype) :: inp
      TYPE(CommonParticleDataType) :: cpd  ! common particle data

      REAL(8) t
      INTEGER nstill(inp%nsteps)

      REAL(8), ALLOCATABLE :: dndt(:),rnsp(:),rnspFIT(:)
      REAL(8), ALLOCATABLE :: ma20(:),ma10(:),ma5(:)
      REAL(8), ALLOCATABLE :: dma20dt(:),dma10dt(:),dma5dt(:)

      INTEGER i
      CHARACTER*255 vrstica

      ALLOCATE(dndt(inp%nsteps))
      ALLOCATE(rnsp(inp%nsteps)) ! relative number of still particles
      ALLOCATE(rnspFIT(inp%nsteps)) ! fitted relative number of still particles
      ALLOCATE(ma20(inp%nsteps))
      ALLOCATE(ma10(inp%nsteps))
      ALLOCATE(ma5(inp%nsteps))
      ALLOCATE(dma20dt(inp%nsteps))
      ALLOCATE(dma10dt(inp%nsteps))
      ALLOCATE(dma5dt(inp%nsteps))


c     calculate relative number of still particles
      DO i=1,inp%nsteps
        rnsp(i)=DBLE(nstill(i))/DBLE(cpd%cnp)
      END DO
      CALL CalDer(rnsp,dndt,inp%nsteps,inp%tstep)
      CALL MovingAverage(rnsp,ma5,inp%nsteps,5)
      CALL CalDer(ma5,dma5dt,inp%nsteps,inp%tstep)
      CALL MovingAverage(rnsp,ma10,inp%nsteps,10)
      CALL CalDer(ma10,dma10dt,inp%nsteps,inp%tstep)
      CALL MovingAverage(rnsp,ma20,inp%nsteps,200)
      CALL CalDer(ma20,dma20dt,inp%nsteps,inp%tstep)

c     fit
      CALL nlfit(io,rnsp,inp%nsteps,inp%tstep,rnspFIT)

      OPEN (io%ns,FILE=io%ns_name,STATUS='UNKNOWN')

c     naslov, variables
      WRITE (io%ns,'(A)')
     & 'VARIABLES = "time", "nstill","rnsp","rnspFIT", "drnsp/dt","ma5","dma5/dt","ma10","dma10/dt","ma200","dma200/dt"'
      WRITE (io%ns,'(A,G16.9,A)') 'ZONE T="B0=',cpd%H0*1.256637061D-6,'"'

C       Output results
      t=0.0D0
      DO i=1,inp%nsteps
        t=t+inp%tstep
        WRITE (vrstica,'(G18.9,I10,9G18.9)')
     & t,nstill(i),rnsp(i),rnspFIT(i),dndt(i),ma5(i),dma5dt(i),ma10(i),dma10dt(i),ma20(i),dma20dt(i)
        CALL sqblnk(io%ns,vrstica)
      END DO

      CLOSE(io%ns)
      DEALLOCATE (dndt,rnsp,rnspFIT,ma20,ma10,ma5,dma20dt,dma10dt,dma5dt)

      END

C -----------------------------------------------------------------------------
      SUBROUTINE TESTnlfit(io)
      INCLUDE 'inc_types.f'

      REAL(8), ALLOCATABLE :: y(:)
      REAL(8), ALLOCATABLE :: yfit(:)
      REAL(8) a,h
      TYPE(IOtype)    :: io
      INTEGER ndata,i
      CHARACTER*255 dummy

      ndata=300000
      h=1.0d0/ndata !0.001D0
      ALLOCATE (yfit(ndata),y(ndata))

      OPEN (UNIT=99,FILE="vns.dat")

      READ(99,*) dummy
      READ(99,*) dummy

      DO i=1,ndata
c        DO j=1,950
c          READ(99,*) dummy
c        END DO
        READ(99,*) a,a,y(i)
      END DO

      CALL nlfit(io,y,ndata,h,yfit)

      WRITE(98,*) 'VARIABLES="x","y","yfit"'
      DO i=1,ndata
        WRITE(98,*) i,y(i),yfit(i)
      END DO


      END




C -----------------------------------------------------------------------------
      SUBROUTINE nlfit(io,y,ndata,h,yfit)
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'
      REAL(8) h
      TYPE(IOtype)    :: io

      INTEGER ndata,ma,nca,i,j,n,model
      REAL(8), ALLOCATABLE :: x(:),sig(:),a(:),covar(:,:),alpha(:,:),ap(:)
      INTEGER, ALLOCATABLE :: ia(:)
      REAL(8) alamda,chisq,y(ndata),yfit(ndata)
      REAL(8) res,res1,sumy,sumdif

      ma=2
      nca=2
      ALLOCATE (x(ndata),sig(ndata))
      ALLOCATE (ia(ma))
      ALLOCATE (a(ma),ap(ma),alpha(nca,nca),covar(nca,nca))

      ia(1)=1
      ia(2)=1

c     iks gre do 1
      DO i=1,ndata
        x(i)=i*h
c        x(i)=DBLE(i)/DBLE(ndata)
        sig(i)=1.0D0
      END DO
      sig(ndata)=0.0000000001D0 ! zadnja tocno

c     zacetna priblizka
      a(1)=y(ndata)
      a(2)=1.0D0

c     poiscem kje narastne na polovico
      DO i=1,ndata
        IF (y(i).GT.y(ndata)*0.5D0) THEN
c          a(2)=-LOG(1-y(i)/y(ndata))/x(i)
          a(2)=LOG(2.0D0)/x(i)
          EXIT
        END IF
      END DO
      WRITE (io%l,'(/A,2F15.8)') "Zacetna priblizka=",a(1),a(2)

      model=5

      alamda=-1.0D0
      res=1.0D0
      n=0
      DO WHILE (res.GT.1.0D-6.AND.n.LT.101)
        ap=a
        n=n+1
        CALL mrqmin(x,y,sig,ndata,a,ia,ma,covar,alpha,nca,chisq,alamda,model)
c        print *,alamda,a(1),a(2),"a",chisq
        res=0.0D0
        res1=0.0D0
        DO j=1,ma
          res=res+(ap(j)-a(j))**2
          res1=res1+a(j)**2
        END DO
        IF (res1.NE.0.0D0) res=res/res1
      END DO
      alamda=0.0D0
      CALL mrqmin(x,y,sig,ndata,a,ia,ma,covar,alpha,nca,chisq,alamda,model)

      WRITE (io%l,'(A,I5,2F15.8)') "Nonlinear fit=",n,a(1),a(2)
      WRITE (io%l,'(A,F15.8)') "hi kvadrat=",chisq  ! enacba 15.5.5 v NR
      WRITE (io%l,'(A,2G18.9)') "covar(1,1),covar(1,2) =",covar(1,1),covar(1,2)
      WRITE (io%l,'(A,2G18.9)') "covar(2,1),covar(2,2) =",covar(2,1),covar(2,2)
      WRITE (io%l,'(A,2G18.9)') "alpha(1,1),alpha(1,2) =",alpha(1,1),alpha(1,2)
      WRITE (io%l,'(A,2G18.9)') "alpha(2,1),alpha(2,2) =",alpha(2,1),alpha(2,2)
c      print *,n,a(1),a(2)

      DO i=1,ndata
        yfit(i)=a(1)*(1.0D0-EXP(-a(2)*x(i)))
      END DO

      sumy=0.0D0
      sumdif=0.0D0
      DO i=1,ndata
        sumy=sumy+y(i)**2
        sumdif=sumdif+(y(i)-yfit(i))**2
      END DO
!     izkaze se da je chisq  = sumdif
      IF (sumy.GT.0.0D0) THEN
        sumdif=sumdif/sumy
        WRITE (io%l,'(A,G18.9)') "SQRT sum (y-yfit)^2 / sum y^2 =",SQRT(sumdif)
      END IF

      END




C -----------------------------------------------------------------------------
      SUBROUTINE MovingAverage(f,ma,n,s)
C
C     naredi moving averaga za s nazaj + moja + s naprej
C     pri s=1 torej u_i=(u_i-1+u_i+u_i+1)/3
C -----------------------------------------------------------------------------
      INTEGER n,i,s,j
      REAL(8) f(n),ma(n),sirina

      sirina=2*s+1

c     zacetni del in koncni del krajsi moving average
      ma=0.0D0
      DO i=1,s
        DO j=i-(i-1),i+(i-1)
          ma(i)=ma(i)+f(j)
        END DO
        ma(i)=ma(i)/(2*i-1)
      END DO
c     delam moving average
      DO i=s+1,n-s
        DO j=i-s,i+s
          ma(i)=ma(i)+f(j)
        END DO
        ma(i)=ma(i)/sirina
      END DO
c     koncni del
      DO i=n-s+1,n
        DO j=i-(n-i),i+(n-i)
          ma(i)=ma(i)+f(j)
        END DO
        ma(i)=ma(i)/(2*(n-i)+1)
      END DO


      END

C -----------------------------------------------------------------------------
      SUBROUTINE CalDer(f,odvod,n,h)
C
C -----------------------------------------------------------------------------
      INTEGER n,i
      REAL(8) h
      REAL(8) f(n),odvod(n)

      DO i=1,n-1
        odvod(i)=( f(i+1)-f(i) ) / h
      END DO
      odvod(n)=( f(n) - f(n-1) ) / h

      END

C -----------------------------------------------------------------------------
      SUBROUTINE CalDerSO(f,odvod,n,h)
C
C -----------------------------------------------------------------------------
      INTEGER n,i
      REAL(8) h
      REAL(8) f(n),odvod(n)

      odvod(1)=( f(2) - f(1) ) / h
      DO i=2,n-1
        odvod(i)=(1.5D0*f(i+1)-2.0D0*f(i)+0.5D0*f(i-1)) / h
      END DO
      odvod(n)=( f(n) - f(n-1) ) / h

      END


C -----------------------------------------------------------------------------
      SUBROUTINE outPD(io,t,part)
C
C     $: writes a zone into results file
C
C -----------------------------------------------------------------------------            
      INCLUDE 'inc_types.f'
   
      TYPE(particleType) :: part  
      INTEGER io,a
      REAL(8) t
      
      IF (part%active.EQ..TRUE.) THEN
        a=part%cell
      ELSE
        a=0
      END IF
      
      WRITE (io,'(12F18.12,2I8)') 
     &                          part%x,part%y,part%z,t,part%st,part%dp,
     &                          part%vx,part%vy,part%vz,
     &                          part%vsx,part%vsy,part%vsz,part%id,a
      
      END

C -----------------------------------------------------------------------------
      SUBROUTINE CopyBIN2TXT(env,io,icount,inp)
C
C     $: copies binary data on particle locations to txt files
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(IOtype)    :: io
      TYPE(penv) :: env
      TYPE(inputtype) :: inp

      REAL(8) t
      INTEGER n,lun,myproc,icount(env%nproc),i,j
      REAL(8) x,y,z,vx,vy,vz,v
      INTEGER :: id       ! identifikacijska stevilka delca
      LOGICAL :: active   ! ali je v toku ali ga je odneslo ven ali se je posedel
      INTEGER :: cell     ! celica v kateri je
      INTEGER :: iPeri    ! v kateri periodi je
      REAL(8) :: St       ! Stokes number
      INTEGER :: StClass       ! Stokes number class
      REAL(8) :: dp       ! premer delca
      REAL(8) :: vsx,vsy,vsz ! settling velocity delca
      REAL(8) :: Rep      ! Particle Reynolds number = dp*f_u0/f_visk
      INTEGER :: ppno     ! Particle processor number, which processor handles the particle


      CHARACTER*60 name
      CHARACTER*255 vrstica
      INTEGER fnblc,lnzc,np,mp
	  INTEGER MaxIcount ! najvecje stevilo time stepov
	  LOGICAL prvic

C     Open binary files
      DO myproc=1,env%nproc
        lun=io%ibin+myproc
        IF (myproc.LT.10) THEN
          WRITE(name,'(A,I1)') 'vreteno.bin.',myproc
        ELSE IF (myproc.LT.100) THEN
          WRITE(name,'(A,I2)') 'vreteno.bin.',myproc
        ELSE IF (myproc.LT.1000) THEN
          WRITE(name,'(A,I3)') 'vreteno.bin.',myproc
        END IF
        OPEN (lun,FILE=TRIM(name),FORM='UNFORMATTED',STATUS='OLD')
        READ (lun) np,mp !nproc,myproc
      END DO
C     open txt files
C     Tecplot
      OPEN (io%trf1,FILE=io%trf1_name,STATUS='UNKNOWN')
      WRITE (io%trf1,'(A)') 'VARIABLES = "x","y","z","vx","vy","vz","v"'

c     poiscem procesor, ki ima najvecje stevilo timestepov
      MaxIcount=-1
    	DO i=1,env%nproc
	      IF (icount(i).GT.MaxIcount) MaxIcount=icount(i)
    	END DO

c     start exporting loop


      DO i=1,MaxIcount
	  prvic=.TRUE.
        DO myproc=1,env%nproc
	    IF (i.LE.icount(myproc)) THEN ! samo ce imam toliko timestepov
            lun=io%ibin+myproc
            READ (lun) t
            READ (lun) n
            IF (prvic.EQ..TRUE.) THEN
              WRITE (io%trf1,'(A10,F10.6,A1)') 'ZONE T="t=',t,'"'
              WRITE (vrstica,'(F16.9)') t
              WRITE (io%trf1,'(A,A,A)') 'AUXDATA time="',vrstica(fnblc(vrstica):lnzc(vrstica)),'"'
	        prvic=.FALSE.
            END IF
            DO j=1,n
              READ (lun) x,y,z,vx,vy,vz,v
              WRITE (vrstica,'(7F18.9)') x,y,z,vx,vy,vz,v
              CALL sqblnk(io%trf1,vrstica)
            END DO
	    END IF
        END DO
      END DO ! time step
C     Close txt files
      CLOSE(io%trf1)

C     final particle informaiton
      OPEN (io%ple,FILE=io%ple_name,STATUS='UNKNOWN')
      OPEN (io%plr,FILE=io%plr_name,STATUS='UNKNOWN')
      WRITE (io%ple,'(A)')
     &'VARIABLES = "x","y","z","t","St","dp","Rep","vx","vy","vz","vsx","vsy","vsz","pn","cell","ppno","iPeri","active"'
      DO myproc=1,env%nproc
        lun=io%ibin+myproc
        READ (lun) t
        READ (lun) n
        DO j=1,n
          READ (lun) x,y,z,vx,vy,vz,vsx,vsy,vsz,st,dp,rep
          READ (lun) id,cell,stclass,ppno,iPeri
          READ (lun) active
          IF (active.EQ..true.) WRITE (io%plr,'(A9,7F18.12,I10)') "PART 4 0 ",x,y,z,st,vx,vy,vz,iPeri
          x=x+(iPeri-1)*inp%Xlength ! dodam periodicnost
          WRITE (vrstica,'(13F15.9,5I6,I1)') x,y,z,t,st,dp,rep,vx,vy,vz,vsx,vsy,vsz,id,cell,ppno,iPeri,active
          CALL sqblnk(io%ple,vrstica)
        END DO
      END DO

      CLOSE (io%plr)
      CLOSE (io%ple)
C     Close binary files
      DO myproc=1,env%nproc
        CLOSE(io%ibin+myproc)
      END DO

      END




C -----------------------------------------------------------------------------
      SUBROUTINE CopyPdensBIN2TXT(env,io,mesh,icount)
C
C     $: copies binary data on particle density to txt files
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(IOtype)    :: io
      TYPE(penv) :: env
      TYPE(meshType)  :: mesh

      REAL(8) :: pdensf(mesh%nnodes)
      REAL(8) :: pdens(mesh%nnodes)

      REAL(8) t
      INTEGER lun,myproc,icount,i,j,k,itp,nic

      CHARACTER*60 name
      CHARACTER*255 vrstica
      CHARACTER*100 zonetitle

      INTEGER fnblc,lnzc,np,mp

C     Open binary files
      DO myproc=1,env%nproc
        lun=io%ipdens+myproc
        IF (myproc.LT.10) THEN
          WRITE(name,'(A,I1)') 'vreteno.pdens.bin.',myproc
        ELSE IF (myproc.LT.100) THEN
          WRITE(name,'(A,I2)') 'vreteno.pdens.bin.',myproc
        ELSE IF (myproc.LT.1000) THEN
          WRITE(name,'(A,I3)') 'vreteno.pdens.bin.',myproc
        END IF
        OPEN (lun,FILE=TRIM(name),FORM='UNFORMATTED',STATUS='OLD')
        READ (lun) np,mp !nproc,myproc
      END DO
C     open txt files
C     Tecplot
      itp=77
      nic=8*mesh%nicell
      OPEN (itp,FILE="vreteno.pdens.dat",STATUS='UNKNOWN')
      WRITE (itp,'(A)') 'VARIABLES = "X", "Y", "Z", "node","pdens"'


c     start exporting loop
      DO k=1,icount
        pdens=0.0D0
        DO myproc=1,env%nproc
          lun=io%ipdens+myproc
          READ (lun) t
          CALL RDvec(lun,mesh%nnodes,pdensf)
          DO j=1,mesh%nnodes
            pdens(j)=pdens(j)+pdensf(j)
          END DO
        END DO
        WRITE (zonetitle,'(F16.9)') t
        IF (k.EQ.1) THEN
          WRITE (itp,'(A,A,A,I8,A,I8)')
     &      'ZONE T="',trim(zonetitle),'",F=FEPOINT, ET=BRICK,N= ', mesh%NNODES, ',E=', nic
          WRITE (itp,'(A,A,A)') 'AUXDATA time="',zonetitle(fnblc(zonetitle):lnzc(zonetitle)),'"'
C         Output results
          DO  i=1,mesh%nnodes
            WRITE (vrstica,'(3G18.9,I8,G18.9)') mesh%x(i,1),mesh%x(i,2),mesh%x(i,3),i,pdens(i)
            CALL sqblnk(itp,vrstica)
          END DO
          CALL TECwIDC(itp,mesh)
        ELSE
          WRITE (itp,'(A8,A,A,I8,A,I8,A)')
     &      'ZONE T="',trim(zonetitle),'",F=FEPOINT, ET=BRICK,N= ', mesh%NNODES, ',E=', nic,
     &      ', VARSHARELIST=([1,2,3,4]=1), CONNECTIVITYSHAREZONE = 1'
          WRITE (itp,'(A,A,A)') 'AUXDATA time="',zonetitle(fnblc(zonetitle):lnzc(zonetitle)),'"'
C         Output results
          DO  i=1,mesh%nnodes
            WRITE (vrstica,'(G18.9)') pdens(i)
            CALL sqblnk(itp,vrstica)
          END DO
        END IF
      END DO
      CLOSE(itp)

C     Close binary files
      DO myproc=1,env%nproc
        CLOSE(io%ipdens+myproc)
      END DO

      END




C -----------------------------------------------------------------------------
      SUBROUTINE OutBINpdens(io,pdens,nnodes,x)
C
C     $: writes particle density data into binary file
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(IOtype)    :: io
      INTEGER nnodes
      REAL(8) pdens(nnodes),x

      WRITE (io%ipdens) x
      CALL wrvec(io%ipdens,nnodes,pdens)

      END


C----------------------------------------------------------------------C
c----------------------------------------------------------------------c
c                                                                      c
      SUBROUTINE rdvec(ifr,nnx,vec)
c                                                                      c
c----------------------------------------------------------------------c
C----------------------------------------------------------------------C
c.......................................................................
c..                                                                   ..
c..   REad VECtor using MAXSIZE chunks                                ..
c..   --   ---                                                        ..
c.......................................................................
      INTEGER ifr,nnx,maxsize,nblock,i,j,k
      REAL*8  vec(nnx)
      PARAMETER (maxsize=8192)

      nblock=INT(nnx/maxsize)
      DO j=1,nblock
        k=maxsize*(j-1)
        READ(ifr) (vec(i),i=k+1,k+maxsize)
      END DO
      k=maxsize*nblock
      READ(ifr) (vec(i),i=k+1,nnx)
      RETURN
      END
C----------------------------------------------------------------------C
c----------------------------------------------------------------------c
c                                                                      c
      SUBROUTINE wrvec(ifr,nnx,vec)
c                                                                      c
c----------------------------------------------------------------------c
C----------------------------------------------------------------------C
c.......................................................................
c..                                                                   ..
c..   WRite VECtor using MAXSIZE chunks                               ..
c..   --    ---                                                       ..
c.......................................................................
      INTEGER ifr,nnx,maxsize,nblock,i,j,k
      REAL*8  vec(nnx)
      PARAMETER (maxsize=8192)
c....
      nblock=INT(nnx/maxsize)
      DO j=1,nblock
        k=maxsize*(j-1)
        WRITE(ifr) (vec(i),i=k+1,k+maxsize)
      END DO
      k=maxsize*nblock
      WRITE(ifr) (vec(i),i=k+1,nnx)
      RETURN
      END


C -----------------------------------------------------------------------------
      SUBROUTINE outBIN(env,x,io,cpd,part)
C
C     $: writes particle data into binary file
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(IOtype)    :: io
      TYPE(CommonParticleDataType) :: cpd  ! common particle data
      TYPE(particleType) :: part(cpd%maxNp) ! seznam delcev
      TYPE(penv) :: env

      INTEGER i,n,lun
      REAL(8) x ! cas

      lun=env%myproc+io%ibin
      WRITE (lun) x ! cas

c     prestejemo koliko jih bom izpisal
      n=0
      DO i=1,cpd%maxNp
        IF ((part(i)%active.EQ..TRUE.).AND.(env%myproc.EQ.part(i)%ppno)) n=n+1
      END DO

      WRITE (lun) n

      DO i=1,cpd%maxNp
        IF (part(i)%active.EQ..TRUE..AND.env%myproc.EQ.part(i)%ppno) THEN
          WRITE (lun) part(i)%x,part(i)%y,part(i)%z,part(i)%vx,part(i)%vy,part(i)%vz,
     &    SQRT(part(i)%vx**2+part(i)%vy**2+part(i)%vz**2)
        END IF
      END DO

      END

C -----------------------------------------------------------------------------
      SUBROUTINE outBINfinal(env,x,io,cpd,part)
C
C     $: writes particle data into binary file (final time step
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(IOtype)    :: io
      TYPE(CommonParticleDataType) :: cpd  ! common particle data
      TYPE(particleType) :: part(cpd%maxNp) ! seznam delcev
      TYPE(penv) :: env

      INTEGER i,n,lun
      REAL(8) x ! cas

      lun=env%myproc+io%ibin

      WRITE (lun) x ! cas

c     prestejemo koliko jih bom izpisal
      n=0
      DO i=1,cpd%maxNp
        IF (env%myproc.EQ.part(i)%ppno) n=n+1
      END DO

      WRITE (lun) n

      DO i=1,cpd%maxNp
        IF (env%myproc.EQ.part(i)%ppno) THEN
          WRITE (lun) part(i)%x,part(i)%y,part(i)%z,
     &                    part(i)%vx,part(i)%vy,part(i)%vz,
     &                    part(i)%vsx,part(i)%vsy,part(i)%vsz,
     &                    part(i)%St,part(i)%dp,part(i)%rep
          WRITE (lun) part(i)%id,part(i)%cell,part(i)%Stclass,part(i)%ppno,part(i)%iPeri
          WRITE (lun) part(i)%active
        END IF
      END DO



      END

C -----------------------------------------------------------------------------
      SUBROUTINE outTRFzone(x,io,inp,cpd,part)
C
C     $: writes a zone into results file
C
C -----------------------------------------------------------------------------            
      INCLUDE 'inc_types.f'
  
      TYPE(IOtype)    :: io  
      TYPE(CommonParticleDataType) :: cpd  ! common particle data
      TYPE(particleType) :: part(cpd%maxNp) ! seznam delcev
      TYPE(inputtype) :: inp
      CHARACTER*255 vrstica   
      INTEGER fnblc,lnzc
      LOGICAL enaktiven

      INTEGER i
      REAL(8) x

c     cfx
      IF (inp%cfxo.EQ.1) THEN
        WRITE (io%trf,'(A10,F10.6,A1)') 'ZONE T="t=',x,'"'
        DO i=1,cpd%maxNp
          IF (part(i)%active.EQ..TRUE.) THEN
            WRITE (io%trf,'(I8,8F15.9)') part(i)%id,part(i)%x,part(i)%y,part(i)%z,x,part(i)%vx,part(i)%vy,part(i)%vz,
     &      SQRT(part(i)%vx**2+part(i)%vy**2+part(i)%vz**2)
          END IF
        END DO
      END IF

c     tecplot      
C     v tecplotu naredi txt objekt in vanj napisi : t=&(AUXZONE[ACTIVEOFFSET=1]:rtime0)s
C     to potem izpise itime za vsako zono svoj.

c     ali je vsaj en aktiven
      enaktiven=.FALSE.
      DO i=1,cpd%maxNp
        IF (part(i)%active.EQ..TRUE.) THEN
          enaktiven=.TRUE.
          EXIT
        END IF
      END DO
      
      IF (enaktiven.EQ..TRUE.) THEN
        WRITE (io%trf1,'(A10,F10.6,A1)') 'ZONE T="t=',x,'"'
        WRITE (vrstica,'(F16.9)') x
        WRITE (io%trf1,'(A,A,A)') 'AUXDATA time="',vrstica(fnblc(vrstica):lnzc(vrstica)),'"'
      
        DO i=1,cpd%maxNp
          IF (part(i)%active.EQ..TRUE.) THEN
            WRITE (vrstica,'(7F15.9)') part(i)%x,part(i)%y,part(i)%z,part(i)%vx,part(i)%vy,part(i)%vz,
     &      SQRT(part(i)%vx**2+part(i)%vy**2+part(i)%vz**2)
            CALL sqblnk(io%trf1,vrstica)
          END IF
        END DO
      END IF

      END


C -----------------------------------------------------------------------------
      SUBROUTINE outLIST(x,io,cpd,part)
C
C     $: writes a particle list into results file
C
C -----------------------------------------------------------------------------            
      INCLUDE 'inc_types.f'
  
      TYPE(IOtype)    :: io  
      TYPE(CommonParticleDataType) :: cpd  ! common particle data
      TYPE(particleType) :: part(cpd%maxNp) ! seznam delcev
      CHARACTER*255 vrstica   

      INTEGER i
      REAL(8) x


c     tecplot      
      DO i=1,cpd%maxNp
        IF (part(i)%active.EQ..TRUE.) THEN
          WRITE (vrstica,'(8F15.9,I8)') part(i)%x,part(i)%y,part(i)%z,part(i)%vx,part(i)%vy,part(i)%vz,
     &      SQRT(part(i)%vx**2+part(i)%vy**2+part(i)%vz**2),x,i
          CALL sqblnk(io%trf2,vrstica)
        END IF
      END DO

      END


C______________________________________________________________________C
      INTEGER FUNCTION fnblc(niz)
C
C      returns the number of the first non blank character in a string
C______________________________________________________________________C
      CHARACTER*(*) niz
      INTEGER i
   
      DO i=1,Len_trim(niz)
        IF (niz(i:i).NE.' ') THEN
          fnblc=i
          RETURN
        END IF
      END DO

c     string containes only blanks
      fnblc=1
C
      RETURN
      END

C______________________________________________________________________C
      INTEGER FUNCTION lnzc(niz)
C
C      returns the number of the last nonzero character after . in a string
C______________________________________________________________________C
      CHARACTER*(*) niz
      INTEGER i,j,pika,lt
    
      lt=len_trim(niz)
      pika=0
      DO i=1,lt
        IF (niz(i:i).EQ.'.') THEN
          pika=i
        END IF
      END DO
      
      IF (pika.EQ.0.OR.lt-pika-1.LT.0) THEN
        lnzc=lt
        RETURN
      END IF

      DO i=0,Lt-pika-2
        j=lt-i
        IF (niz(j:j).NE.'0') THEN
          lnzc=j
          RETURN
        END IF
      END DO

      lnzc=pika+1
C
      RETURN
      END


C -----------------------------------------------------------------------------
      
      SUBROUTINE SetUpIO(env,io,inp)
C
C     $: sets up input and output file names and numbers
C
C -----------------------------------------------------------------------------       
      INCLUDE 'inc_types.f'
      TYPE(IOtype) io
      TYPE(inputtype) inp
      TYPE(penv) :: env
      CHARACTER(50) hostn      
            
c     log file
      IF (env%myproc.EQ.1) THEN
        io%l=11
        io%l_name='vreteno.log'
        OPEN (io%l,FILE=io%l_name,STATUS='UNKNOWN')
        WRITE (io%l,'(/6A/)') trim(inp%IDname)," ",trim(inp%IDversion),", ",trim(inp%IDdate),"."
        WRITE (io%l,'(A)') inp%StartTime
        CALL hostnam(hostn)
        WRITE (io%l,'(A,A/)') "Running on: ",hostn
      END IF
      
      
c     input file      
      io%inp=12
      io%inp_name='vreteno.inp'

c     mesh neighbour list file      
      io%neil=16
      io%neil_name='vreteno.neil'      
      
c     mesh file
      io%mesh=14      

c     error/warning file      
      IF (env%myproc.EQ.1) THEN
        io%errwarn=13
        io%errwarn_name='vreteno.err'
        OPEN (io%errwarn,FILE=io%errwarn_name,STATUS='UNKNOWN')
        WRITE (io%errwarn,'(/6A/)') trim(inp%IDname)," ",trim(inp%IDversion),", ",trim(inp%IDdate),"."
      END IF
      
c     fluid results data file
      io%redf=15      
      
c     tekst results file
      io%trf=17
      io%trf_name='vreteno.dat'

c     tekst results file
      io%trf1=20
      io%trf1_name='vreteno.tec.dat'

c     tekst results file
      io%trf2=21
      io%trf2_name='vreteno.list.dat'

c     particle list file
      io%pl=18
      io%pl_name='vreteno.plist.start.dat'
      io%ple=19
      io%ple_name='vreteno.plist.end.dat'
      io%plr=23
      io%plr_name='vreteno.restart'

c     list of visited cells
      io%lvc=22
      io%lvc_name='vreteno.lvc.dat'

c     kjer magnetna sila dominira
      io%mfdom=33
      io%mfdom_name='vreteno.mfdom.dat'
c     statistika sil
      io%fs=34
      io%fs_name='vreteno.fstat.dat'
c     minimalna razdalja do zice
      io%md=35
      io%md_name='vreteno.mdist.dat'
c     stevilo mirujocih delcev
      io%ns=36
      io%ns_name='vreteno.nstill.dat'


c     binary results file
c     integrals file
      io%ibin=88 ! pazi vzame tudi naprej
      IF (env%myproc.LT.10) THEN
        WRITE(io%ibin_name,'(A,I1)') 'vreteno.bin.',env%myproc
      ELSE IF (env%myproc.LT.100) THEN
        WRITE(io%ibin_name,'(A,I2)') 'vreteno.bin.',env%myproc
      ELSE IF (env%myproc.LT.1000) THEN
        WRITE(io%ibin_name,'(A,I3)') 'vreteno.bin.',env%myproc
      END IF

c     binary pdens results file
      io%ipdens=31
      IF (env%myproc.LT.10) THEN
        WRITE(io%ipdens_name,'(A,I1)') 'vreteno.pdens.bin.',env%myproc
      ELSE IF (env%myproc.LT.100) THEN
        WRITE(io%ipdens_name,'(A,I2)') 'vreteno.pdens.bin.',env%myproc
      ELSE IF (env%myproc.LT.1000) THEN
        WRITE(io%ipdens_name,'(A,I3)') 'vreteno.pdens.bin.',env%myproc
      END IF

c     monitor forces results file
      io%imf=32
      IF (env%myproc.LT.10) THEN
        WRITE(io%imf_name,'(A,I1,A4)') 'vreteno.monf.',env%myproc,".dat"
      ELSE IF (env%myproc.LT.100) THEN
        WRITE(io%imf_name,'(A,I2,A4)') 'vreteno.monf.',env%myproc,".dat"
      ELSE IF (env%myproc.LT.1000) THEN
        WRITE(io%imf_name,'(A,I3,A4)') 'vreteno.monf.',env%myproc,".dat"
      END IF


     
      END       
      
C______________________________________________________________________C  

      SUBROUTINE WarnErr(io,inp,WorE,routine,text,ierr)
C                       
C     $: Writes warning or error to error file
C______________________________________________________________________C
      INCLUDE 'inc_types.f'
      TYPE(IOtype) io
      TYPE(inputtype) inp
      
      CHARACTER*(*) routine,text
      CHARACTER*255 msgtype
      INTEGER WorE,ierr

C
C       Message, warning or error
C
      IF (WorE.EQ.0) THEN
      WRITE(msgtype,'(A7,A4,A,A4,A)') 
     &      "Message"," :: ",trim(routine)," :: ",trim(text)
      ELSE IF (WorE.EQ.5) THEN
       WRITE(msgtype,'(A7,A4,A,A4,A,I8)') 
     &      "Message"," :: ",trim(routine)," :: ",trim(text),ierr
      ELSE IF (WorE.EQ.1) THEN
       WRITE(msgtype,'(A7,A4,A,A4,A)') 
     &      "Warning"," :: ",trim(routine)," :: ",trim(text)
      ELSE IF (WorE.EQ.2) THEN
       WRITE(msgtype,'(A7,A4,A,A4,A)') 
     &      "Error  "," :: ",trim(routine)," :: ",trim(text)
      ELSE IF (WorE.EQ.3) THEN
       WRITE(msgtype,'(A7,A4,A,A4,A,I8)') 
     &      "Warning"," :: ",trim(routine)," :: ",trim(text),ierr
      ELSE IF (WorE.EQ.4) THEN
          WRITE(msgtype,'(A7,A4,A,A4,A,I8)') 
     &      "Error  "," :: ",trim(routine)," :: ",trim(text),ierr
       ELSE
         msgtype="Unkonwn"
        END IF
C
C       write to error file
C   
        WRITE(io%errwarn,'(A)') trim(msgtype)
C
C       stop if error ocured
C     
      IF (WorE.EQ.2.OR.WorE.EQ.4) THEN
        CALL StopProgram(io,inp,1)
      END IF

      END
      
C______________________________________________________________________C
C______________________________________________________________________C
      SUBROUTINE WrIMat(mat,nrow,ncol,io)
C        __    _      ___
C        Write Single Matrix
C______________________________________________________________________C
C______________________________________________________________________C
      INTEGER nrow,ncol,io,j
      INTEGER mat(nrow,ncol)

      DO j=1,ncol
        CALL wrIvec(io,nrow,mat(1,j))
      END DO

      END
C______________________________________________________________________C
C______________________________________________________________________C
      SUBROUTINE RdIMat(mat,nrow,ncol,io)
C        _  _ _      ___
C        Read Single Matrix
C______________________________________________________________________C
C______________________________________________________________________C
      INTEGER nrow,ncol,io,j
      INTEGER mat(nrow,ncol)

      DO j=1,ncol
        CALL rdIvec(io,nrow,mat(1,j))
      END DO

      END

C----------------------------------------------------------------------C
c----------------------------------------------------------------------c
c                                                                      c
      SUBROUTINE rdIvec(ifr,nnx,vec)
c                                                                      c
c----------------------------------------------------------------------c
C----------------------------------------------------------------------C
c.......................................................................
c..                                                                   ..
c..   REad VECtor using MAXSIZE chunks                                ..
c..   --   ---                                                        ..
c.......................................................................
      INTEGER ifr,nnx,maxsize,nblock,i,j,k
      INTEGER  vec(nnx)
      PARAMETER (maxsize=8192)

      nblock=INT(nnx/maxsize)
      DO j=1,nblock
        k=maxsize*(j-1)
        READ(ifr) (vec(i),i=k+1,k+maxsize)
      END DO
      k=maxsize*nblock
      READ(ifr) (vec(i),i=k+1,nnx)
      RETURN
      END
C----------------------------------------------------------------------C
c----------------------------------------------------------------------c
c                                                                      c
      SUBROUTINE wrIvec(ifr,nnx,vec)
c                                                                      c
c----------------------------------------------------------------------c
C----------------------------------------------------------------------C
c.......................................................................
c..                                                                   ..
c..   WRite VECtor using MAXSIZE chunks                               ..
c..   --    ---                                                       ..
c.......................................................................
      INTEGER ifr,nnx,maxsize,nblock,i,j,k
      INTEGER  vec(nnx)
      PARAMETER (maxsize=8192)
c....
      nblock=INT(nnx/maxsize)
      DO j=1,nblock
        k=maxsize*(j-1)
        WRITE(ifr) (vec(i),i=k+1,k+maxsize)
      END DO
      k=maxsize*nblock
      WRITE(ifr) (vec(i),i=k+1,nnx)
      RETURN
      END
            

C______________________________________________________________________C
      SUBROUTINE TECwIDC(lun,mesh)
C
C     Tecplot write cell conectivity                                   C
C______________________________________________________________________C
      INCLUDE 'inc_types.f'
      TYPE(meshType) mesh
C....
      CHARACTER*200 vrstica
      INTEGER i,lun
      
      INTEGER a1(8),a2(8),a3(8),a4(8),a5(8),a6(8),a7(8),a8(8)
      
      DATA a1/6,18,23,14,17,26,27,22/
      DATA a2/18,7,15,23,26,19,24,27/
      DATA a3/14,23,10,2,22,27,21,9/
      DATA a4/23,15,3,10,27,24,11,21/
      DATA a5/17,26,27,22,5,20,25,13/
      DATA a6/26,19,24,27,20,8,16,25/
      DATA a7/22,27,21,9,13,25,12,1/
      DATA a8/27,24,11,21,25,16,4,12/
C
C     Write cell connectivity
C
      IF (mesh%npoc.EQ.9) THEN
        DO i=1,mesh%nicell
          WRITE(vrstica,*) mesh%idc(i,1),mesh%idc(i,2),mesh%idc(i,9),mesh%idc(i,8)
          CALL sqblnk(lun,vrstica)
          WRITE(vrstica,*) mesh%idc(i,2),mesh%idc(i,3),mesh%idc(i,4),mesh%idc(i,9)
          CALL sqblnk(lun,vrstica)          
          WRITE(vrstica,*) mesh%idc(i,8),mesh%idc(i,9),mesh%idc(i,6),mesh%idc(i,7)
          CALL sqblnk(lun,vrstica)          
          WRITE(vrstica,*) mesh%idc(i,9),mesh%idc(i,4),mesh%idc(i,5),mesh%idc(i,6)
          CALL sqblnk(lun,vrstica)          
        END DO
      ELSE IF (mesh%npoc.EQ.4) THEN
        DO i=1,mesh%nicell
          WRITE(vrstica,*) mesh%idc(i,1),mesh%idc(i,2),mesh%idc(i,3),mesh%idc(i,4)
          CALL sqblnk(lun,vrstica)          
        END DO
      ELSE IF (mesh%npoc.EQ.27) THEN
        DO i=1,mesh%nicell
          WRITE(vrstica,*) mesh%idc(i,a1(1)),mesh%idc(i,a1(2)),mesh%idc(i,a1(3)),mesh%idc(i,a1(4))
     &                    ,mesh%idc(i,a1(5)),mesh%idc(i,a1(6)),mesh%idc(i,a1(7)),mesh%idc(i,a1(8))
          CALL sqblnk(lun,vrstica)
          WRITE(vrstica,*) mesh%idc(i,a2(1)),mesh%idc(i,a2(2)),mesh%idc(i,a2(3)),mesh%idc(i,a2(4))
     &                    ,mesh%idc(i,a2(5)),mesh%idc(i,a2(6)),mesh%idc(i,a2(7)),mesh%idc(i,a2(8))
          CALL sqblnk(lun,vrstica)
          WRITE(vrstica,*) mesh%idc(i,a3(1)),mesh%idc(i,a3(2)),mesh%idc(i,a3(3)),mesh%idc(i,a3(4))
     &                    ,mesh%idc(i,a3(5)),mesh%idc(i,a3(6)),mesh%idc(i,a3(7)),mesh%idc(i,a3(8))
          CALL sqblnk(lun,vrstica)
          WRITE(vrstica,*) mesh%idc(i,a4(1)),mesh%idc(i,a4(2)),mesh%idc(i,a4(3)),mesh%idc(i,a4(4))
     &                    ,mesh%idc(i,a4(5)),mesh%idc(i,a4(6)),mesh%idc(i,a4(7)),mesh%idc(i,a4(8))
          CALL sqblnk(lun,vrstica)
          WRITE(vrstica,*) mesh%idc(i,a5(1)),mesh%idc(i,a5(2)),mesh%idc(i,a5(3)),mesh%idc(i,a5(4))
     &                    ,mesh%idc(i,a5(5)),mesh%idc(i,a5(6)),mesh%idc(i,a5(7)),mesh%idc(i,a5(8))
          CALL sqblnk(lun,vrstica)
          WRITE(vrstica,*) mesh%idc(i,a6(1)),mesh%idc(i,a6(2)),mesh%idc(i,a6(3)),mesh%idc(i,a6(4))
     &                    ,mesh%idc(i,a6(5)),mesh%idc(i,a6(6)),mesh%idc(i,a6(7)),mesh%idc(i,a6(8))
          CALL sqblnk(lun,vrstica)
          WRITE(vrstica,*) mesh%idc(i,a7(1)),mesh%idc(i,a7(2)),mesh%idc(i,a7(3)),mesh%idc(i,a7(4))
     &                    ,mesh%idc(i,a7(5)),mesh%idc(i,a7(6)),mesh%idc(i,a7(7)),mesh%idc(i,a7(8))
          CALL sqblnk(lun,vrstica)
          WRITE(vrstica,*) mesh%idc(i,a8(1)),mesh%idc(i,a8(2)),mesh%idc(i,a8(3)),mesh%idc(i,a8(4))
     &                    ,mesh%idc(i,a8(5)),mesh%idc(i,a8(6)),mesh%idc(i,a8(7)),mesh%idc(i,a8(8))
          CALL sqblnk(lun,vrstica)

        END DO

      END IF

      END    
      
C *********************************************************************
C **                                                                 **
C ** Squeeze multiple blanks into single blank                       **
C **                                                                 **
C *********************************************************************
C---------------------------------------------------------------------C
      SUBROUTINE sqblnk(lun,line)
C---------------------------------------------------------------------C
      CHARACTER*(*) line
      LOGICAL flag
      INTEGER i,ii,j,lnblnk,lun

      j=1
      flag=.FALSE.
      DO i=1,LNBLNK(line)+1
        IF (line(i:i).NE.' ' .AND. .NOT.flag) THEN
          flag=.TRUE.
          ii=i
        ELSE IF (line(i:i).EQ.' ' .AND. flag) THEN
          flag=.FALSE.
          line(j:j+i-ii)=line(ii:i)
          j=j+i-ii+1
        END IF
      END DO
      WRITE(lun,'(A)') line(1:j-2)
      RETURN
      END
      INTEGER FUNCTION LNBLNK (string)
C
C     LNBLNK returns the index of the last non-blank character in string
C
      CHARACTER*(*) string
      INTEGER i

      DO i=LEN(string),1,-1
        IF (string(i:i).NE.' ') THEN
          lnblnk=i
          RETURN
        END IF
      END DO
      lnblnk=0
      RETURN
      END             
