C -----------------------------------------------------------------------------
      SUBROUTINE HeliBC(part)
C
C     $: when particle hits a wall it is copied over the diagonal
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(particleType) :: part
      REAL(8) dum1

      IF (part%x.GT.1.0D0) part%x=part%x-1.0D0
      IF (part%x.LT.0.0D0) part%x=part%x+1.0D0

c     spodaj
      IF (part%z.LT.0.0D0.AND.part%y.LT.0.5D0) THEN
        dum1=part%y
        part%y=1.0D0+part%z
        part%z=1.0D0-dum1

        dum1=part%vy
        part%vy=part%vz
        part%vz=dum1    
      END IF

      IF (part%z.LT.0.0D0.AND.part%y.GE.0.5D0) THEN
        part%y=1.0D0-part%y
        part%z=1.0D0+part%z
        
        part%vy=-part%vy
        part%vz=-part%vz    
      END IF

c     zgoraj
      IF (part%z.GT.1.0D0.AND.part%y.LT.0.5D0) THEN
        part%y=1.0D0-part%y
        part%z=part%z-1.0D0

        part%vy=-part%vy
        part%vz=-part%vz    
      END IF

      IF (part%z.GT.1.0D0.AND.part%y.GE.0.5D0) THEN
        dum1=part%y
        part%y=part%z-1.0D0
        part%z=1.0D0-dum1

        dum1=part%vy
        part%vy=part%vz
        part%vz=dum1    

      END IF

c     levo
      IF (part%y.LT.0.0D0.AND.part%z.LT.0.5D0) THEN
        part%y=1.0D0+part%y
        part%z=1.0D0-part%z

        part%vy=-part%vy
        part%vz=-part%vz    
      END IF

      IF (part%y.LT.0.0D0.AND.part%z.GE.0.5D0) THEN
        dum1=part%y
        part%y=part%z
        part%z=-dum1

        dum1=part%vy
        part%vy=-part%vz
        part%vz=-dum1    

      END IF

c     desno
      IF (part%y.GT.1.0D0.AND.part%z.LT.0.5D0) THEN
        dum1=part%y
        part%y=part%z
        part%z=2.0D0-dum1

        dum1=part%vy
        part%vy=-part%vz
        part%vz=-dum1    
      END IF

      IF (part%y.GT.1.0D0.AND.part%z.GE.0.5D0) THEN
        part%y=part%y-1.0D0
        part%z=1.0D0-part%z

        part%vy=-part%vy
        part%vz=-part%vz    
      END IF


      
      END


C -----------------------------------------------------------------------------
      SUBROUTINE BounceBC(part)
C
C     $: when particle hits a wall it bounces
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(particleType) :: part

      IF (part%x.GT.1.0D0) part%x=part%x-1.0D0
      IF (part%x.LT.0.0D0) part%x=part%x+1.0D0
      
      IF (part%y.GT.1.0D0) THEN
        part%y=1.0D0-(part%y-1.0D0)
        part%vy=-part%vy
      END IF
      IF (part%y.LT.0.0D0) THEN
        part%y=-part%y
        part%vy=-part%vy
      END IF

      IF (part%z.GT.1.0D0) THEN
        part%z=1.0D0-(part%z-1.0D0)
        part%vz=-part%vz
      END IF
      IF (part%z.LT.0.0D0) THEN
        part%z=-part%z
        part%vz=-part%vz
      END IF

      END
      
C -----------------------------------------------------------------------------
      SUBROUTINE GenHeliFlow(mesh,fluid,cpd)
C
C     $: generira heli flow field
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
      TYPE(fluidType) :: fluid  ! hitrostno, vrtincno in T polje     
      TYPE(CommonParticleDataType) :: cpd  ! common particle data   
      
      REAL(8) x,y,z
      REAL(8) stiriPi,w,y0,z0
      INTEGER node

      ALLOCATE (fluid%vx(mesh%nnodes))
      ALLOCATE (fluid%vy(mesh%nnodes))
      ALLOCATE (fluid%vz(mesh%nnodes))
      ALLOCATE (fluid%wx(mesh%nnodes))
      ALLOCATE (fluid%wy(mesh%nnodes))
      ALLOCATE (fluid%wz(mesh%nnodes))
      ALLOCATE (fluid%T(mesh%nnodes))
      ALLOCATE (fluid%pdens(mesh%nnodes))

      ALLOCATE (fluid%gradVx(mesh%nnodes,3))
      ALLOCATE (fluid%gradVy(mesh%nnodes,3))
      ALLOCATE (fluid%gradVz(mesh%nnodes,3))
      ALLOCATE (fluid%dvxdt(mesh%nnodes))
      ALLOCATE (fluid%dvydt(mesh%nnodes))
      ALLOCATE (fluid%dvzdt(mesh%nnodes))

      stiriPi=16.0D0*ATAN(1.0D0)
      w=1.0D0
      z0=0.5D0
      y0=0.5D0
      
      

      DO node=1,mesh%nnodes
        x=mesh%x(node,1)
        y=mesh%x(node,2)
        z=mesh%x(node,3)

        fluid%vx(node)=cpd%f_u0*(1.0D0+0.5D0*sin(stiriPi*x))
        fluid%vy(node)=-(z-z0)*w
        fluid%vz(node)=+(y-y0)*w
        fluid%wx(node)=2.0D0*w
        fluid%wy(node)=0.0D0
        fluid%wz(node)=0.0D0
        fluid%T(node)=0.0D0
      END DO

C     set up derivatives 
      CALL setGRAD(mesh,fluid%vx,fluid%gradVx)
      CALL setGRAD(mesh,fluid%vy,fluid%gradVy)
      CALL setGRAD(mesh,fluid%vz,fluid%gradVz)
      fluid%dvxdt=0.0D0
      fluid%dvydt=0.0D0
      fluid%dvzdt=0.0D0

      END
      
C -----------------------------------------------------------------------------
      SUBROUTINE wHeliFlowTec(mesh,fluid)
C
C     $: writes heli flow to disk
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
      TYPE(fluidType) :: fluid  ! hitrostno, vrtincno in T polje
      CHARACTER*200 vrstica
      
      INTEGER itp,nic,i
      
      itp=77
      nic=8*mesh%nicell
      OPEN (itp,FILE="heli.dat",STATUS='UNKNOWN')
      
      WRITE (itp,'(A)') 'VARIABLES = "X", "Y", "Z", "node","vx","vy","vz","wx","wy","wz","T"'
      
      WRITE (itp,'(A,I8,A,I8)') 
     &      'ZONE T="heli flow field",F=FEPOINT, ET=BRICK,N= ', mesh%NNODES, ',E=', nic
     
C       Output results        
        DO  i=1,mesh%nnodes
          WRITE (vrstica,'(3G18.9,I8,7G18.9)') 
     &        mesh%x(i,1),mesh%x(i,2),mesh%x(i,3),i,
     &        fluid%vx(i),fluid%vy(i),fluid%vz(i),
     &        fluid%wx(i),fluid%wy(i),fluid%wz(i),
     &        fluid%T(i)
     
          CALL sqblnk(itp,vrstica)           
        END DO
        CALL TECwIDC(itp,mesh)
      
      CLOSE(itp)
      
      END   
      

C -----------------------------------------------------------------------------
      SUBROUTINE GenCellFlow(mesh,xpi,fluid)
C
C     $: generira cellular flow field
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
      TYPE(fluidType) :: fluid  ! hitrostno, vrtincno in T polje        
      
      REAL(8) x,y,z
      REAL(8) dvaPi,xpi
      INTEGER node

      ALLOCATE (fluid%vx(mesh%nnodes))
      ALLOCATE (fluid%vy(mesh%nnodes))
      ALLOCATE (fluid%vz(mesh%nnodes))
      ALLOCATE (fluid%wx(mesh%nnodes))
      ALLOCATE (fluid%wy(mesh%nnodes))
      ALLOCATE (fluid%wz(mesh%nnodes))
      ALLOCATE (fluid%T(mesh%nnodes))
      ALLOCATE (fluid%pdens(mesh%nnodes))

      ALLOCATE (fluid%gradVx(mesh%nnodes,3))
      ALLOCATE (fluid%gradVy(mesh%nnodes,3))
      ALLOCATE (fluid%gradVz(mesh%nnodes,3))
      ALLOCATE (fluid%dvxdt(mesh%nnodes))
      ALLOCATE (fluid%dvydt(mesh%nnodes))
      ALLOCATE (fluid%dvzdt(mesh%nnodes))

      dvaPi=xpi*4.0D0*ATAN(1.0D0)

      DO node=1,mesh%nnodes
        x=mesh%x(node,1)
        y=mesh%x(node,2)
        z=mesh%x(node,3)

c        fluid%vx(node)=-cos(dvapi*x)*sin(dvapi*z)
        fluid%vx(node)=sin(dvapi*x)*cos(dvapi*z)        
        fluid%vy(node)=0.0D0
        fluid%vz(node)=-cos(dvapi*x)*sin(dvapi*z)
        fluid%wx(node)=0.0D0
        fluid%wy(node)=2.0D0*dvapi*sin(dvapi*x)*sin(dvapi*z)
        fluid%wz(node)=0.0D0
        fluid%T(node)=0.0D0
        
        fluid%gradVx(node,1)=dvapi*cos(dvapi*x)*cos(dvapi*z)
        fluid%gradVx(node,2)=0.0D0
        fluid%gradVx(node,3)=-dvapi*sin(dvapi*x)*sin(dvapi*z)

        fluid%gradVy(node,1)=0.0D0
        fluid%gradVy(node,2)=0.0D0
        fluid%gradVy(node,3)=0.0D0

        fluid%gradVz(node,1)=dvapi*sin(dvapi*x)*sin(dvapi*z)
        fluid%gradVz(node,2)=0.0D0
        fluid%gradVz(node,3)=-dvapi*cos(dvapi*x)*cos(dvapi*z)

      END DO

C     set up derivatives 
c      CALL setGRAD(mesh,fluid%vx,fluid%gradVx)
c      CALL setGRAD(mesh,fluid%vy,fluid%gradVy)
c      CALL setGRAD(mesh,fluid%vz,fluid%gradVz)

      fluid%dvxdt=0.0D0
      fluid%dvydt=0.0D0
      fluid%dvzdt=0.0D0

      END
      
C -----------------------------------------------------------------------------
      SUBROUTINE wCellFlowTec(mesh,fluid)
C
C     $: generira cellular flow field
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
      TYPE(fluidType) :: fluid  ! hitrostno, vrtincno in T polje
      CHARACTER*200 vrstica
      
      INTEGER itp,nic,i
      
      itp=77
      nic=8*mesh%nicell
      OPEN (itp,FILE="cell.dat",STATUS='UNKNOWN')
      
      WRITE (itp,'(A)') 'VARIABLES = "X", "Y", "Z", "node","vx","vy","vz","wx","wy","wz","T"'
      
      WRITE (itp,'(A,I8,A,I8)') 
     &      'ZONE T="cell flow field",F=FEPOINT, ET=BRICK,N= ', mesh%NNODES, ',E=', nic
     
C       Output results        
        DO  i=1,mesh%nnodes
          WRITE (vrstica,'(3G18.9,I8,7G18.9)') 
     &        mesh%x(i,1),mesh%x(i,2),mesh%x(i,3),i,
     &        fluid%vx(i),fluid%vy(i),fluid%vz(i),
     &        fluid%wx(i),fluid%wy(i),fluid%wz(i),
     &        fluid%T(i)
     
          CALL sqblnk(itp,vrstica)           
        END DO
        CALL TECwIDC(itp,mesh)
      
      CLOSE(itp)
      
      END   
      
      
C -----------------------------------------------------------------------------
      SUBROUTINE Gen1pt(mesh,fluid)
C
C     $: generira u=2*x
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
      TYPE(fluidType) :: fluid  ! hitrostno, vrtincno in T polje        
      
      REAL(8) x,y,z
      INTEGER node

      ALLOCATE (fluid%vx(mesh%nnodes))
      ALLOCATE (fluid%vy(mesh%nnodes))
      ALLOCATE (fluid%vz(mesh%nnodes))
      ALLOCATE (fluid%wx(mesh%nnodes))
      ALLOCATE (fluid%wy(mesh%nnodes))
      ALLOCATE (fluid%wz(mesh%nnodes))
      ALLOCATE (fluid%T(mesh%nnodes))
      ALLOCATE (fluid%pdens(mesh%nnodes))

      ALLOCATE (fluid%gradVx(mesh%nnodes,3))
      ALLOCATE (fluid%gradVy(mesh%nnodes,3))
      ALLOCATE (fluid%gradVz(mesh%nnodes,3))
      ALLOCATE (fluid%dvxdt(mesh%nnodes))
      ALLOCATE (fluid%dvydt(mesh%nnodes))
      ALLOCATE (fluid%dvzdt(mesh%nnodes))


      DO node=1,mesh%nnodes
        x=mesh%x(node,1)
        y=mesh%x(node,2)
        z=mesh%x(node,3)

        fluid%vx(node)=2.0D0*x
        fluid%vy(node)=0.0D0
        fluid%vz(node)=0.0D0
        fluid%wx(node)=0.0D0
        fluid%wy(node)=0.0D0
        fluid%wz(node)=0.0D0
        fluid%T(node)=0.0D0
      END DO

C     set up derivatives 
      CALL setGRAD(mesh,fluid%vx,fluid%gradVx)
      CALL setGRAD(mesh,fluid%vy,fluid%gradVy)
      CALL setGRAD(mesh,fluid%vz,fluid%gradVz)
      fluid%dvxdt=0.0D0
      fluid%dvydt=0.0D0
      fluid%dvzdt=0.0D0

      END                 
