C -----------------------------------------------------------------------------
      SUBROUTINE ReadMeshGid3D(io,inp,mesh,gp)
C
C     $: reads mesh
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'
      
      TYPE(meshType) :: mesh
      TYPE(gausstype) :: gp      
      TYPE(IOtype) :: io
      TYPE(InputType) inp
        
      INTEGER lun,i,j,k,kk,dn,be
      CHARACTER KeyWord*4,OneLine*255,tmp*255 
      
      lun=io%mesh
      
      OPEN (lun,FILE=trim(mesh%fullname),ERR=10,STATUS='OLD',SHARED)

c
C***  READ VERSION NUMBER OR TITLE OF GEO INPUT FILE :
c
      CALL rOneTL(lun,OneLine)
      READ (OneLine,*) KeyWord
      mesh%version=TRIM(OneLine(6:len_trim(OneLine)))

      CALL rOneTL(lun,OneLine)
      mesh%title=OneLine(1:len_trim(OneLine))
c
C***  READ MESH SIZE :
c
      CALL rOneTL(lun,OneLine)
      READ(OneLine,*) mesh%nbelem,mesh%nicell,mesh%nnodes,mesh%nbnodes,mesh%npob,mesh%npoc,mesh%npx
      IF (mesh%npx.NE.3) THEN
        CALL WarnErr(io,inp,4,"ReadMeshGid3D","only npx=3 supported!",0)
      ELSE
        READ(OneLine,*) mesh%nbelem,mesh%nicell,mesh%nnodes,mesh%nbnodes,mesh%npob,mesh%npoc,mesh%npx
c     &                  ,mesh%nx,mesh%ny,mesh%simx,mesh%simy,mesh%nofc,mesh%nofw
      END IF
c      mesh%nbnodm1=mesh%nbnodes-1
c      mesh%nnmnb=mesh%nnodes-mesh%nbnodes
c      mesh%nxpny2=(mesh%nx+mesh%ny)/2

      ALLOCATE (mesh%ibc(mesh%nbelem,mesh%npob),mesh%idc(mesh%nicell,mesh%npoc))
      ALLOCATE (mesh%x(mesh%nnodes,mesh%npx))
      ALLOCATE (mesh%bewn(mesh%nbelem))
      ALLOCATE (mesh%gbn(mesh%nbnodes),mesh%lbn(mesh%nnodes))
c      ALLOCATE (mesh%ukode(mesh%nnodes))
      
c      ALLOCATE (mesh%area(mesh%nicell),mesh%aleng(mesh%nbelem))
c      ALLOCATE (mesh%un(mesh%nbnodes,mesh%npx),mesh%ut(mesh%nbnodes,mesh%npx))

c      ALLOCATE (mesh%iwn(mesh%nbnodes),mesh%iwc(mesh%nofw,mesh%npob))
c      ALLOCATE (mesh%icn(mesh%nofw))
        
c      ALLOCATE (mesh%isw(mesh%nofw))    
c      ALLOCATE (mesh%kobc(mesh%nbnodes))            
c
C____ READ IN MESH POINT COORDINATES:
c
      mesh%id=0.0D00
      DO i=1,mesh%nnodes
        CALL rOneTL(lun,OneLine)
        READ (OneLine,*) k,(mesh%x(k,j),j=1,mesh%npx)
        DO j=1,mesh%npx
          mesh%id=mesh%id+SQRT(ABS(mesh%x(k,j)))
        END DO        
      END DO
C
C____ READ BOUNDARY ELEMENT CONNECTIVITY  :
C
      DO i=1,mesh%nbelem
        CALL rOneTL(lun,OneLine)
        READ (OneLine,*) k,(mesh%ibc(k,j),j=1,mesh%npob)
      END DO                  
C
C____ READ IN INTERNAL CELL CONNECTIVITY  :
C
      DO i=1,mesh%nicell
        CALL rOneTL(lun,OneLine)
        READ (OneLine,*) k,(mesh%idc(k,j),j=1,mesh%npoc)
      END DO
C
C____ FORM BOUNDARY NODES VECTOR
C       
      k=0
      mesh%gbn=0
      mesh%lbn=0
      DO i=1,mesh%nbelem
        DO j=1,mesh%npob
          dn=mesh%ibc(i,j)
          DO kk=1,k
            IF (mesh%gbn(kk).EQ.dn) GOTO 100
          END DO
          k=k+1
          mesh%lbn(dn)=k
          mesh%gbn(k)=dn
100      END DO
      END DO
C
C____ READ BOUNDARY ELEMENT WALL NUMBER :
C
      DO i=1,mesh%nbelem
        CALL rOneTL(lun,OneLine)
        READ (OneLine,*) be,mesh%bewn(be)
      END DO
C
C____ NUMBER OF CELLS TIMES NUMBER OF POINTS IN A CELL
C          
      mesh%nicnpoc=mesh%nicell*mesh%npoc

      CLOSE (lun)
C     Calculate Cell volume
      ALLOCATE (mesh%vol(mesh%nicell))
      CALL MeshVolume(gp,mesh)      

C     Swithc for domination of magnetic force
      ALLOCATE (mesh%mfdom(mesh%nicell))
      mesh%mfdom=0
      

C
C     Calculate normal in every boundary point
C
c      CALL SourcePointNormals(mesh)
C
C     find number of walls
C
      mesh%nofw=MAXVAL(mesh%bewn)
C
C     nnodes -nbnodes
C
      mesh%nnmnb=mesh%nnodes - mesh%nbnodes      
c
c     seznam samo obnocnih vozlisc DNL(nnmnb), obratno IDNL(nnodes)
c
      ALLOCATE (mesh%dnl(mesh%nnmnb),mesh%idnl(mesh%nnodes))
      j=0
      mesh%dnl=0
      mesh%idnl=0

      DO i=1,mesh%nnodes
        IF (mesh%lbn(i).EQ.0) THEN ! je obmocno vozlisce 
          j=j+1
          mesh%dnl(j)=i
          mesh%idnl(i)=j
        END IF
      END DO

      RETURN
      
10    continue ! error when opening mesh
      WRITE (tmp,'(A,A)') "could not open mesh: ",trim(mesh%fullname)
      CALL WarnErr(io,inp,2,"ReadMeshGid3D",trim(tmp),0)

      END
      
      
C -----------------------------------------------------------------------------
      SUBROUTINE MakeNeighList(mesh)
C
C     $: naredi seznam sosedov vsake celice
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'
      
      TYPE(meshType) :: mesh
      
      INTEGER i,j,k,l
      LOGICAL jeSosed,imamNaSeznamu
      

      mesh%neil=0
           
      DO i=1,mesh%nicell
        DO j=1,mesh%nicell
          IF (i.NE.j) THEN
c           celica j je soseda celice i, ce je katerikoli node v obeh celicah
            jeSosed=.FALSE.
            DO k=1,mesh%npoc
              DO l=1,mesh%npoc
                IF (mesh%idc(i,k).EQ.mesh%idc(j,l)) THEN
                  jeSosed=.TRUE.
                  GOTO 10
                END IF
              END DO  !l
            END DO  !k
10          CONTINUE            
c            print *,i,j,k,l,jesosed
            IF (jeSosed) THEN
c             poiscem, ce ga ze imam na seznamu
              imamNaSeznamu=.FALSE.
              DO l=1,mesh%neil(i,27)
                IF (mesh%neil(i,l).EQ.j) THEN
                  imamNaSeznamu=.TRUE.
                  GOTO 20
                END IF
              END DO
20            CONTINUE              
              IF (imamNaSeznamu.EQ..FALSE.) THEN
                mesh%neil(i,27)=mesh%neil(i,27)+1
                mesh%neil(i,mesh%neil(i,27))=j
              END IF
            END IF            
          END IF       
        END DO  !j
      END DO !i
      
      
      END
      
