
C -----------------------------------------------------------------------------
      SUBROUTINE CavitydpBC(cpd,part)
C
C     $: Elasticen odboj od vseh sten z upostevanjem premera delca
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(particleType) :: part
      TYPE(CommonParticleDataType) :: cpd  ! common particle data
      REAL(8) bdpd

c     brezdimenzijski polmer delca
      bdpd=part%dp/cpd%f_L/2.0D0

c     spodaj
      IF (part%z.LT.bdpd) THEN
        part%z=2.0D0*bdpd-part%z
        part%vz=-part%vz      
      END IF      

c     zgoraj
      IF (part%z.GT.1.0D0-bdpd) THEN
        part%z=2.0D0-2.0D0*bdpd-part%z
        part%vz=-part%vz
        part%vx=1.0D0  ! lid ga prime in povlece      
      END IF      

c     levo
      IF (part%x.LT.bdpd) THEN
        part%x=2.0D0*bdpd-part%x
        part%vx=-part%vx      
      END IF      

c     desno
      IF (part%x.GT.1.0D0-bdpd) THEN
        part%x=2.0D0-2.0D0*bdpd-part%x
        part%vx=-part%vx      
      END IF      

c     spredaj
      IF (part%y.LT.bdpd) THEN
        part%y=2.0D0*bdpd-part%y
        part%vy=-part%vy      
      END IF      

c     zadaj
      IF (part%y.GT.1.0D0-bdpd) THEN
        part%y=2.0D0-2.0D0*bdpd-part%y
        part%vy=-part%vy      
      END IF      

      END




C -----------------------------------------------------------------------------
      SUBROUTINE CavityBC(part)
C
C     $: Elasticen odboj od vseh sten
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(particleType) :: part

c     spodaj
      IF (part%z.LT.0.0D0) THEN
        part%z=-part%z
        part%vz=-part%vz      
      END IF      

c     zgoraj
      IF (part%z.GT.1.0D0) THEN
        part%z=2.0D0-part%z
        part%vz=-part%vz
        part%vx=1.0D0  ! lid ga prime in povlece      
      END IF      

c     levo
      IF (part%x.LT.0.0D0) THEN
        part%x=-part%x
        part%vx=-part%vx      
      END IF      

c     desno
      IF (part%x.GT.1.0D0) THEN
        part%x=2.0D0-part%x
        part%vx=-part%vx      
      END IF      

c     spredaj
      IF (part%y.LT.0.0D0) THEN
        part%y=-part%y
        part%vy=-part%vy      
      END IF      

c     zadaj
      IF (part%y.GT.1.0D0) THEN
        part%y=2.0D0-part%y
        part%vy=-part%vy      
      END IF      

      END



C -----------------------------------------------------------------------------
      SUBROUTINE ChannelBC(part,nPeri,Xlength)
C
C     $: Elasticen odboj od vseh sten
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'
      INTEGER nPeri
      REAL(8) Xlength

      TYPE(particleType) :: part

c     spodaj
c      IF (part%z.LT.0.0D0) THEN
c        part%z=-part%z
c        part%vz=-part%vz
c      END IF

c     zgoraj
c      IF (part%z.GT.1.0D0) THEN
c        part%z=2.0D0-part%z
c        part%vz=-part%vz
c      END IF

c     spredaj
      IF (part%y.LT.0.0D0) THEN
        part%y=-part%y
        part%vy=-part%vy
      END IF

c     zadaj
      IF (part%y.GT.1.0D0) THEN
        part%y=2.0D0-part%y
        part%vy=-part%vy
      END IF

c     letel ven
      IF (part%x.GT.Xlength) THEN ! periodicno nazaj
        IF (part%iPeri.LT.nPeri) THEN
          part%x=part%x-Xlength
          part%iPeri=part%iPeri+1
        END IF
      END IF


      END



C -----------------------------------------------------------------------------
      SUBROUTINE cal_dvdt(pn,part,cpd,fap,y,dydx)
C
C     $: calculate right hand side of the momentum equation, dv/dt
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'
  
      TYPE(fluidAtPartType) :: fap ! tekocina na mestu delca, fluid at the location of the particle
      TYPE(CommonParticleDataType) :: cpd  ! common particle data
      TYPE(particleType) :: part(cpd%maxNp) ! seznam delcev, list of particles
      INTEGER pn ! particle number
      
      REAL(8) dvdt(3),dydx(6)
      REAL(8) cpv(3),y(6) ! current particle velocity, to je tista, ki jo 
                     ! runge kutta popravlja, zacetna je v fap
                     ! ko je runge kutta skozi, potem fap%vx=cpv(1)
      REAL(8) hpct ! heavy particle correction term
      REAL(8) avd  ! absolute value of veloctiy difference

c     y
      cpv(1)=y(2) ! vx
      cpv(2)=y(4) ! vy
      cpv(3)=y(6) ! vz

c     heavy particle correction
      IF (cpd%hepc.EQ.1) THEN
        avd=SQRT( (fap%vx-cpv(1))**2 + (fap%vy-cpv(2))**2 + (fap%vz-cpv(2))**2 ) 
        hpct=0.15D0*(part(pn)%rep*avd)**0.687D0
      ELSE
        hpct=0.0D0
      END IF

C     smer x      
      dvdt(1)=cpd%A/part(pn)%st*
     &                         (part(pn)%vsx+   ! settling velocity x
     &                         (fap%vx-cpv(1))*(1.0D0+hpct))+  ! drag
     &        1.5D0*cpd%R*fap%dvxdt+
     &        cpd%R*(
     &               (fap%vx+0.5D0*cpv(1))*fap%gradVx(1)+
     &               (fap%vy+0.5D0*cpv(2))*fap%gradVx(2)+
     &               (fap%vz+0.5D0*cpv(3))*fap%gradVx(3))+
     &        cpd%A*fap%nhhX ! magnetic force

C     smer y      
      dvdt(2)=cpd%A/part(pn)%st*
     &                         (part(pn)%vsy+   ! settling velocity y
     &                         (fap%vy-cpv(2))*(1.0D0+hpct))+  ! drag
     &        1.5D0*cpd%R*fap%dvydt+
     &        cpd%R*(
     &               (fap%vx+0.5D0*cpv(1))*fap%gradVy(1)+
     &               (fap%vy+0.5D0*cpv(2))*fap%gradVy(2)+
     &               (fap%vz+0.5D0*cpv(3))*fap%gradVy(3))+
     &        cpd%A*fap%nhhY ! magnetic force

C     smer z
      dvdt(3)=cpd%A/part(pn)%st*
     &                         (part(pn)%vsz+   ! settling velocity z
     &                         (fap%vz-cpv(3))*(1.0D0+hpct))+  ! drag
     &        1.5D0*cpd%R*fap%dvzdt+
     &        cpd%R*(
     &               (fap%vx+0.5D0*cpv(1))*fap%gradVz(1)+
     &               (fap%vy+0.5D0*cpv(2))*fap%gradVz(2)+
     &               (fap%vz+0.5D0*cpv(3))*fap%gradVz(3))+
     &        cpd%A*fap%nhhZ ! magnetic force

      dydx(1)=y(2)    ! dx / dt = vx
      dydx(2)=dvdt(1) ! dvx / dt = ax
      dydx(3)=y(4)    ! dy / dt = vy
      dydx(4)=dvdt(2) ! dvy / dt = ay
      dydx(5)=y(6)    ! dz / dt = vz
      dydx(6)=dvdt(3) ! dvz / dt = az

      END

C -----------------------------------------------------------------------------
      SUBROUTINE mon_forces(pn,part,cpd,fap,y,time,io,mesh,mfpid,fstat,dydx)
C
C     $: monitor forces on indivirual particles
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
      TYPE(ForceStatisticsType) :: fstat
      TYPE(fluidAtPartType) :: fap ! tekocina na mestu delca
      TYPE(CommonParticleDataType) :: cpd  ! common particle data
      TYPE(particleType) :: part(cpd%maxNp) ! seznam delcev
      TYPE(IOtype)    :: io
      INTEGER pn ! particle number
      INTEGER mfpid ! delec za katerega delam izpis
      REAL(8) vsotasil

      REAL(8) dydx(6)
      REAL(8) cpv(3),y(6) ! current particle velocity, to je tista, ki jo
                     ! runge kutta popravlja, zacetna je v fap
                     ! ko je runge kutta skozi, potem fap%vx=cpv(1)
      REAL(8) cpa(3) ! current particle acceleration
      REAL(8) hpct ! heavy particle correction term
      REAL(8) avd  ! absolute value of veloctiy difference
      REAL(8) time,gb,drag,adpc,mag,aabs,vabs

c      current particle velocity
      cpv(1)=y(2) ! vx
      cpv(2)=y(4) ! vy
      cpv(3)=y(6) ! vz
c      current particle acceleration
      cpa(1)=dydx(2) ! ax
      cpa(2)=dydx(4) ! ay
      cpa(3)=dydx(6) ! az

      vabs=SQRT(cpv(1)**2+cpv(2)**2+cpv(3)**2)
      aabs=SQRT(cpa(1)**2+cpa(2)**2+cpa(3)**2)

c     heavy particle correction
      avd=SQRT( (fap%vx-cpv(1))**2 + (fap%vy-cpv(2))**2 + (fap%vz-cpv(2))**2 )
      IF (cpd%hepc.EQ.1) THEN
        hpct=0.15D0*(part(pn)%rep*avd)**0.687D0
      ELSE
        hpct=0.0D0
      END IF

        gb=cpd%A/part(pn)%st*SQRT(part(pn)%vsx**2+part(pn)%vsy**2+part(pn)%vsz**2)
        drag=cpd%A/part(pn)%st*avd*(1.0D0+hpct)
        ! added mass and pressure correction
        adpc=1.5D0*cpd%R*SQRT(fap%dvxdt**2+fap%dvydt**2+fap%dvzdt**2)+
     &       cpd%R*SQRT(
     &               ((fap%vx+0.5D0*cpv(1))*fap%gradVx(1)+
     &               (fap%vy+0.5D0*cpv(2))*fap%gradVx(2)+
     &               (fap%vz+0.5D0*cpv(3))*fap%gradVx(3))**2+
     &               ((fap%vx+0.5D0*cpv(1))*fap%gradVy(1)+
     &               (fap%vy+0.5D0*cpv(2))*fap%gradVy(2)+
     &               (fap%vz+0.5D0*cpv(3))*fap%gradVy(3))**2+
     &               ((fap%vx+0.5D0*cpv(1))*fap%gradVz(1)+
     &               (fap%vy+0.5D0*cpv(2))*fap%gradVz(2)+
     &               (fap%vz+0.5D0*cpv(3))*fap%gradVz(3))**2)
        mag=cpd%A*SQRT(fap%nhhX**2+fap%nhhY**2+fap%nhhZ**2)

c     vsak procesor zapise time trace sil za svoj delec
c     each processor writesi ts own trace
      IF (mfpid.EQ.pn) THEN
        WRITE (io%imf,'(9(E18.9,1X))') time,gb,drag,adpc,mag,drag/gb,adpc/gb,mag/gb,avd
      END IF

c     vsota sil, sum of forces
      vsotasil=gb+drag+adpc+mag
      fstat%gb(part(pn)%cell)=fstat%gb(part(pn)%cell)+gb/vsotasil
      fstat%drag(part(pn)%cell)=fstat%drag(part(pn)%cell)+drag/vsotasil
      fstat%adpc(part(pn)%cell)=fstat%adpc(part(pn)%cell)+adpc/vsotasil
      fstat%mag(part(pn)%cell)=fstat%mag(part(pn)%cell)+mag/vsotasil
c     hitrosti in pospesek, velocity and acceleration
      fstat%vabs(part(pn)%cell)=fstat%vabs(part(pn)%cell)+vabs
      fstat%aabs(part(pn)%cell)=fstat%aabs(part(pn)%cell)+aabs
c     stevilo delcev, ki je prispevalo podatek za izbrano celico v mrezi
      fstat%n(part(pn)%cell)=fstat%n(part(pn)%cell)+1

c      area where magnetic force dominates
      IF (mag.GT.gb.AND.mag.GT.drag.AND.mag.GT.adpc) THEN
        mesh%mfdom(part(pn)%cell)=1
      END IF

      END


C -----------------------------------------------------------------------------
      SUBROUTINE GetFFFap(mesh,fluid,nhh,fap,part,cpd,ipart,ierr)
C
C     $: get fluid properties at the location of the ipart-th particle
C        ierr=0 particle was found, fluid properties are in fap
C        ierr=1 particles was NOT found anywhere in the mesh
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
      TYPE(fluidType) :: fluid  ! hitrostno, vrtincno in T polje    
      TYPE(fluidAtPartType) :: fap ! tekocina na mestu delca
      TYPE(CommonParticleDataType) :: cpd  ! common particle data
      TYPE(particleType) :: part(cpd%maxNp) ! seznam delcev       
      REAL(8) nhh(mesh%nnodes,3)
      
      REAL(8) x,y,z,xi,eta,zeta
      REAL(8) fi(mesh%npoc) ! shape functions
      INTEGER ierr,ipart,cell,i,j,node
      
      x=part(ipart)%x
      y=part(ipart)%y
      z=part(ipart)%z
c     init flow field at the location of the particle
      fap%vx=0.0D0
      fap%vy=0.0D0
      fap%vz=0.0D0                
      fap%wx=0.0D0
      fap%wy=0.0D0
      fap%wz=0.0D0               
      fap%T =0.0D0
      fap%gradVx=0.0D0
      fap%gradVy=0.0D0
      fap%gradVz=0.0D0
      fap%dvxdt=0.0D0
      fap%dvydt=0.0D0
      fap%dvzdt=0.0D0
c     magnetic part nabla (H\codt H)
      fap%nhhX=0.0D0
      fap%nhhY=0.0D0
      fap%nhhZ=0.0D0
c     init
      xi=0.0D0
      eta=0.0D0
      zeta=0.0D0      


c     naprej pogledamo ce je ostal v isti celici
c     is the particle located in the same mesh cell?
      cell=part(ipart)%cell
      CALL tcfp_cross(mesh,cell,x,y,z,ierr)
      IF (ierr.EQ.0) GOTO 40

c     nato pogledamo v sosednje celice
c     it the particle in neighbouring cells?
      DO i=1,mesh%neil(part(ipart)%cell,27)
        cell=mesh%neil(part(ipart)%cell,i)
        CALL tcfp_cross(mesh,cell,x,y,z,ierr)
        IF (ierr.EQ.0) GOTO 40
      END DO        

c     nazadnje gremo cez vse (do tega ne bi smelo nikoli priti)
c     oziroma to se zgodi takrat, ki je delec izven obmocja
c     finally, we check each cell
      DO cell=1,mesh%nicell
        CALL tcfp_cross(mesh,cell,x,y,z,ierr)
        IF (ierr.EQ.0) GOTO 40
      END DO

C     Nisem ga nasel, torej ga ni nikjer v mrezi -> sel v zid??
c     Particle was not found in any mesh cell
      RETURN
      
40    CONTINUE      
c     remember the cell number in which the particle was found
      part(ipart)%cell=cell     
    
     
c     naprej pogledamo ce je v celici, ki smo jo nasli zgoraj
c     calculate local particle coordinates (xi,eta,zeta) in mesh cell
      cell=part(ipart)%cell
      CALL tcfp(mesh,cell,xi,eta,zeta,x,y,z,ierr)
      IF (ierr.EQ.0) GOTO 100

c     (this should never happen)
c     (do tega ne bi smelo nikoli priti)
      cpd%swg=cpd%swg+1    
c     nato pogledamo v sosednje celice
c     look into neighbouring cells
      DO i=1,mesh%neil(part(ipart)%cell,27)
        cell=mesh%neil(part(ipart)%cell,i)
        CALL tcfp(mesh,cell,xi,eta,zeta,x,y,z,ierr)
        IF (ierr.EQ.0) GOTO 100
      END DO        

c     nazadnje gremo cez vse (do tega ne bi smelo nikoli priti)
c     oziroma to se zgodi takrat, ki je delec izven obmocja
c     look in all cells
      DO cell=1,mesh%nicell
        CALL tcfp(mesh,cell,xi,eta,zeta,x,y,z,ierr)
        IF (ierr.EQ.0) GOTO 100
      END DO


C     Nisem ga nasel, torej ga ni nikjer v mrezi -> sel v zid??
c     Paticle was not found
      RETURN

100   CONTINUE
c     delec je bil najden v celici "cell" s koordiantami xi,eta,zeta
c     particle was found in CELL having local coordinates XI,ETA,ZETA
      part(ipart)%cell=cell
c     calculate shape functions      
      CALL cshape27(xi,eta,zeta,fi,mesh%npoc)

c     interpoliram lastnosi tekocine na mestu delca v fap
c     interpolate fluid parameters
      DO i=1,mesh%npoc
        node=mesh%idc(cell,i)
        fap%vx=fap%vx+fluid%vx(node)*fi(i)
        fap%vy=fap%vy+fluid%vy(node)*fi(i)
        fap%vz=fap%vz+fluid%vz(node)*fi(i)                
        fap%wx=fap%wx+fluid%wx(node)*fi(i)
        fap%wy=fap%wy+fluid%wy(node)*fi(i)
        fap%wz=fap%wz+fluid%wz(node)*fi(i)                
        fap%T =fap%T+ fluid%T(node)*fi(i)
        DO j=1,3
          fap%gradVx(j)=fap%gradVx(j)+fluid%gradVx(node,j)*fi(i)
          fap%gradVy(j)=fap%gradVy(j)+fluid%gradVy(node,j)*fi(i)
          fap%gradVz(j)=fap%gradVz(j)+fluid%gradVz(node,j)*fi(i)
        END DO
        fap%dvxdt=fap%dvxdt+fluid%dvxdt(node)*fi(i)
        fap%dvydt=fap%dvydt+fluid%dvydt(node)*fi(i)
        fap%dvzdt=fap%dvzdt+fluid%dvzdt(node)*fi(i)

        fap%nhhx=fap%nhhx+nhh(node,1)*fi(i)
        fap%nhhy=fap%nhhy+nhh(node,2)*fi(i)
        fap%nhhz=fap%nhhz+nhh(node,3)*fi(i)
c       particle density  0.1714677 je minimalna maximalna vrednost fi; fi(2/3,2/3,2/3)
        if (fi(i).gt.0.1714677D0) fluid%pdens(node)=fluid%pdens(node)+fi(i)
      END DO            

      END 


C -----------------------------------------------------------------------------
      SUBROUTINE tcfp_cross(mesh,ic,x,y,z,ierr)
C
C     $: preveri ali je v tej celici delec
C        check if particle is in a mesh cell (ic), using cross products
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
      REAL(8) x,y,z
      INTEGER ic,ierr

      REAL(8) x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4
      REAL(8) x5,y5,z5,x6,y6,z6,x7,y7,z7,x8,y8,z8
          
      
          x1=mesh%x(mesh%idc(ic,6),1)
          y1=mesh%x(mesh%idc(ic,6),2)
          z1=mesh%x(mesh%idc(ic,6),3)          
          x2=mesh%x(mesh%idc(ic,7),1)
          y2=mesh%x(mesh%idc(ic,7),2)
          z2=mesh%x(mesh%idc(ic,7),3)          
          x3=mesh%x(mesh%idc(ic,2),1)
          y3=mesh%x(mesh%idc(ic,2),2)
          z3=mesh%x(mesh%idc(ic,2),3)          
          x4=mesh%x(mesh%idc(ic,3),1)
          y4=mesh%x(mesh%idc(ic,3),2)
          z4=mesh%x(mesh%idc(ic,3),3)          
          x5=mesh%x(mesh%idc(ic,5),1)
          y5=mesh%x(mesh%idc(ic,5),2)
          z5=mesh%x(mesh%idc(ic,5),3)          
          x6=mesh%x(mesh%idc(ic,8),1)
          y6=mesh%x(mesh%idc(ic,8),2)
          z6=mesh%x(mesh%idc(ic,8),3)          
          x7=mesh%x(mesh%idc(ic,1),1)
          y7=mesh%x(mesh%idc(ic,1),2)
          z7=mesh%x(mesh%idc(ic,1),3)          
          x8=mesh%x(mesh%idc(ic,4),1)
          y8=mesh%x(mesh%idc(ic,4),2)
          z8=mesh%x(mesh%idc(ic,4),3)  
        
          
      CALL cipic(x,y,z,x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4,
     &                   x5,y5,z5,x6,y6,z6,x7,y7,z7,x8,y8,z8,ierr)          
         

      END



C -----------------------------------------------------------------------------
      SUBROUTINE tcfp(mesh,ic,xi,eta,zeta,x,y,z,ierr)
C
C     $: preveri ali je v tej celici delec
C        find local coordinates of the particle (xi,eta,zeta)
C        if particle is not in cell "ic", ierr not zero
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
      REAL(8) xi,eta,zeta,x,y,z
      INTEGER ic,ierr

      REAL(8) x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4
      REAL(8) x5,y5,z5,x6,y6,z6,x7,y7,z7,x8,y8,z8
          
      
          x1=mesh%x(mesh%idc(ic,6),1)
          y1=mesh%x(mesh%idc(ic,6),2)
          z1=mesh%x(mesh%idc(ic,6),3)          
          x2=mesh%x(mesh%idc(ic,7),1)
          y2=mesh%x(mesh%idc(ic,7),2)
          z2=mesh%x(mesh%idc(ic,7),3)          
          x3=mesh%x(mesh%idc(ic,2),1)
          y3=mesh%x(mesh%idc(ic,2),2)
          z3=mesh%x(mesh%idc(ic,2),3)          
          x4=mesh%x(mesh%idc(ic,3),1)
          y4=mesh%x(mesh%idc(ic,3),2)
          z4=mesh%x(mesh%idc(ic,3),3)          
          x5=mesh%x(mesh%idc(ic,5),1)
          y5=mesh%x(mesh%idc(ic,5),2)
          z5=mesh%x(mesh%idc(ic,5),3)          
          x6=mesh%x(mesh%idc(ic,8),1)
          y6=mesh%x(mesh%idc(ic,8),2)
          z6=mesh%x(mesh%idc(ic,8),3)          
          x7=mesh%x(mesh%idc(ic,1),1)
          y7=mesh%x(mesh%idc(ic,1),2)
          z7=mesh%x(mesh%idc(ic,1),3)          
          x8=mesh%x(mesh%idc(ic,4),1)
          y8=mesh%x(mesh%idc(ic,4),2)
          z8=mesh%x(mesh%idc(ic,4),3)  

          
      CALL kks2lks(x,y,z,x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4,
     &                   x5,y5,z5,x6,y6,z6,x7,y7,z7,x8,y8,z8,xi,eta,zeta,ierr)          
        
      END

C -----------------------------------------------------------------------------
      SUBROUTINE ReadResultsDataFile(mesh,io,inp,fluid)
C
C     $: reads Tritok results data file
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
      TYPE(IOtype)    :: io
      TYPE(inputtype) :: inp      
      TYPE(fluidType) :: fluid  ! hitrostno, vrtincno in T polje        
      
      CHARACTER OneLine*255,tmp*255 
      REAL(8) dummy,vx,vy,vz,wx,wy,wz,T
      INTEGER node,i

      ALLOCATE (fluid%vx(mesh%nnodes))
      ALLOCATE (fluid%vy(mesh%nnodes))
      ALLOCATE (fluid%vz(mesh%nnodes))
      ALLOCATE (fluid%wx(mesh%nnodes))
      ALLOCATE (fluid%wy(mesh%nnodes))
      ALLOCATE (fluid%wz(mesh%nnodes))
      ALLOCATE (fluid%T(mesh%nnodes))
      ALLOCATE (fluid%pdens(mesh%nnodes))

      ALLOCATE (fluid%gradVx(mesh%nnodes,3))
      ALLOCATE (fluid%gradVy(mesh%nnodes,3))
      ALLOCATE (fluid%gradVz(mesh%nnodes,3))
      ALLOCATE (fluid%dvxdt(mesh%nnodes))
      ALLOCATE (fluid%dvydt(mesh%nnodes))
      ALLOCATE (fluid%dvzdt(mesh%nnodes))

      OPEN  (io%redf,FILE=trim(inp%redf),STATUS='OLD',ERR=100)
10    CALL rOneTL(io%redf,OneLine)
      DO WHILE (OneLine(1:4).NE.'ZONE')
        GOTO 10
      END DO

C     se AUXDATA
11    CALL rOneTL(io%redf,OneLine)
      IF (OneLine(1:4).EQ.'AUXD') GOTO 11

c     prvega sem ze preberal
        READ (OneLine,*) dummy,dummy,dummy,node,vx,vy,vz,wx,wy,wz,T

        fluid%vx(node)=vx
        fluid%vy(node)=vy
        fluid%vz(node)=vz
        fluid%wx(node)=wx
        fluid%wy(node)=wy
        fluid%wz(node)=wz
        fluid%T(node)=T

c     ostali
      DO i=2,mesh%nnodes
        CALL rOneTL(io%redf,OneLine)
        READ (OneLine,*) dummy,dummy,dummy,node,vx,vy,vz,wx,wy,wz,T

        fluid%vx(node)=vx
        fluid%vy(node)=vy
        fluid%vz(node)=vz
        fluid%wx(node)=wx
        fluid%wy(node)=wy
        fluid%wz(node)=wz
        fluid%T(node)=T
      END DO

C     set up derivatives 
      CALL setGRAD(mesh,fluid%vx,fluid%gradVx)
      CALL setGRAD(mesh,fluid%vy,fluid%gradVy)
      CALL setGRAD(mesh,fluid%vz,fluid%gradVz)

      fluid%dvxdt=0.0D0
      fluid%dvydt=0.0D0
      fluid%dvzdt=0.0D0

      RETURN
      
100   CONTINUE ! error when opening mesh
      WRITE (tmp,'(A,A)') "Could not open flow results data file: ",trim(inp%redf)
      CALL WarnErr(io,inp,2,"ReadResultsDataFile",trim(tmp),0)      

      END

C -----------------------------------------------------------------------------
      LOGICAL FUNCTION CheckKdaj(tstep,kdaj)
      INTEGER kdaj,tstep
      LOGICAL add
      
      add=.FALSE.
      IF (kdaj.EQ.0.AND.tstep.EQ.0) THEN
        add=.TRUE.
      END IF
      IF (kdaj.GT.0) THEN
        IF (MOD(tstep,kdaj).EQ.0) THEN
         add=.TRUE.
       END IF
      END IF
      
      CheckKdaj=add
      
      RETURN
      END
      
      

C -----------------------------------------------------------------------------
      SUBROUTINE SetUpParticles(env,mesh,fluid,nhh,part,cpd,tstep,ppno)
C
C     $: postavi zacetne polazaje delcev
C        inserts particles into the flow
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
      TYPE(CommonParticleDataType) :: cpd  ! common particle data
      TYPE(fluidType) :: fluid  ! hitrostno, vrtincno in T polje  
      TYPE(particleType) :: part(cpd%maxNp) ! seznam delcev   
      TYPE(penv) :: env
      REAL(8) nhh(mesh%nnodes,3)
      
      REAL(8) xi,eta,zeta,x,y,z,vx,vy,vz
      INTEGER rndcell,i,j
      LOGICAL CheckKdaj
      
      INTEGER model,kdaj,cnp,tstep
      INTEGER ppno,iPeri ! particle processor number
      REAL(8) st

      DO i=1,cpd%npartdesc
      
        READ(cpd%partdesc(i),*) model
      
        IF (model.EQ.0) THEN
          READ(cpd%partdesc(i),*) model,kdaj,x,y,z,st
          IF (CheckKdaj(tstep,kdaj).EQ..TRUE.) THEN
            cpd%cnp=cpd%cnp+1
            cpd%lid=cpd%lid+1
            CALL AddParticle(env,mesh,fluid,nhh,cpd,part,cpd%cnp,cpd%lid,x,y,z,1,st,ppno)
          END IF
    
        ELSE IF (model.EQ.1) THEN
          READ(cpd%partdesc(i),*) model,kdaj,cnp,st
          IF (CheckKdaj(tstep,kdaj).EQ..TRUE.) THEN
            DO j=1,cnp
              CALL rndIno(rndcell,mesh%nicell) ! izbere nakljucno celico
              CALL rndXiEtaZeta(xi,eta,zeta)   ! izbere nakljucno lokacijo znotraj celice        
              CALL ppos(mesh,rndcell,xi,eta,zeta,x,y,z) ! izracuna lokacijo delca
              cpd%cnp=cpd%cnp+1
              cpd%lid=cpd%lid+1
              CALL AddParticle(env,mesh,fluid,nhh,cpd,part,cpd%cnp,cpd%lid,x,y,z,rndcell,st,ppno)
            END DO
          END IF

        ELSE IF (model.EQ.2) THEN  ! v vsako celico en delec
          READ(cpd%partdesc(i),*) model,kdaj,st
          IF (CheckKdaj(tstep,kdaj).EQ..TRUE.) THEN
            DO j=1,mesh%nicell
              CALL ppos(mesh,j,0.0D0,0.0D0,0.0D0,x,y,z) ! izracuna lokacijo delca
              cpd%cnp=cpd%cnp+1
              cpd%lid=cpd%lid+1
              CALL AddParticle(env,mesh,fluid,nhh,cpd,part,cpd%cnp,cpd%lid,x,y,z,j,st,ppno)
            END DO
          END IF

        ELSE IF (model.EQ.3) THEN ! znana tudi hitrost delca (restart)
          READ(cpd%partdesc(i),*) model,kdaj,x,y,z,st,vx,vy,vz
          IF (CheckKdaj(tstep,kdaj).EQ..TRUE.) THEN
            cpd%cnp=cpd%cnp+1
            cpd%lid=cpd%lid+1
            CALL AddParticleV(env,mesh,fluid,nhh,cpd,part,cpd%cnp,cpd%lid,x,y,z,vx,vy,vz,1,st,ppno,1)
          END IF

        ELSE IF (model.EQ.4) THEN ! znana tudi hitrost delca (restart) + periodicen problem
          READ(cpd%partdesc(i),*) model,kdaj,x,y,z,st,vx,vy,vz,iPeri
          IF (CheckKdaj(tstep,kdaj).EQ..TRUE.) THEN
            cpd%cnp=cpd%cnp+1
            cpd%lid=cpd%lid+1
            CALL AddParticleV(env,mesh,fluid,nhh,cpd,part,cpd%cnp,cpd%lid,x,y,z,vx,vy,vz,1,st,ppno,iPeri)
          END IF
                    
        ELSE
          WRITE (*,*) "Unknown particle model!"
        END IF
      
      END DO
      
      END

C -----------------------------------------------------------------------------
      SUBROUTINE AddParticle(env,mesh,fluid,nhh,cpd,part,i,id,x,y,z,cell,st,ppno)
C
C     $: doda delec, add a single particle into the flow
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'
      TYPE(meshType)  :: mesh
      TYPE(CommonParticleDataType) :: cpd  ! common particle data      
      TYPE(particleType) :: part(cpd%maxNp) ! seznam delcev
      TYPE(fluidType) :: fluid  ! hitrostno, vrtincno in T polje  
      TYPE(fluidAtPartType) :: fap ! tekocina na mestu delca
      TYPE(penv) :: env
      REAL(8) nhh(mesh%nnodes,3)
      
      INTEGER i,id,ierr,cell,ppno
      REAL(8) x,y,z,st

c     na katerem procesorju
      CALL GetNextPartNo(ppno,env%nproc)
      part(i)%ppno=ppno

c     samo ce je moj
      IF (env%myproc.EQ.part(i)%ppno) THEN

        part(i)%id=id
        part(i)%x=x
        part(i)%y=y
        part(i)%z=z
        part(i)%cell=cell
        part(i)%st=st
        part(i)%iPeri=1  ! zacne v prvi periodi

c       IZJEME
        IF (cpd%simt.EQ.2) THEN  ! cellular flow
          part(i)%y=0.02D0 !y  ! SAMO ZA CELL FLOW
        END IF
        IF (cpd%simt.EQ.6) THEN  ! channel flow
          part(i)%x=0.0D0  ! SAMO ZA TOK V KANALU
        END IF

        CALL GetFFFap(mesh,fluid,nhh,fap,part,cpd,i,ierr)
        IF (ierr.EQ.0) THEN
c         set particle active, its velocity equal to fluid velocity
          part(i)%active=.TRUE.
          part(i)%still=.FALSE.
          part(i)%vx=fap%vx
          part(i)%vy=fap%vy
          part(i)%vz=fap%vz
        ELSE
          part(i)%active=.FALSE.
        END IF
c       premer
        part(i)%dp=cpd%f_l*SQRT(18.0D0*cpd%f_rho/cpd%p_rho*part(i)%st/cpd%f_re)
c       settling velocity
        part(i)%vsx=part(i)%dp**2/18.0D0/cpd%f_visk/cpd%f_u0*cpd%g0*cpd%gx*(cpd%p_rho/cpd%f_rho-1.0D0)
        part(i)%vsy=part(i)%dp**2/18.0D0/cpd%f_visk/cpd%f_u0*cpd%g0*cpd%gy*(cpd%p_rho/cpd%f_rho-1.0D0)
        part(i)%vsz=part(i)%dp**2/18.0D0/cpd%f_visk/cpd%f_u0*cpd%g0*cpd%gz*(cpd%p_rho/cpd%f_rho-1.0D0)
c       particle Reynolds number
        part(i)%rep=part(i)%dp*cpd%f_u0/cpd%f_visk
      END IF
      
      END


C -----------------------------------------------------------------------------
      SUBROUTINE AddParticleV(env,mesh,fluid,nhh,cpd,part,i,id,x,y,z,vx,vy,vz,cell,st,ppno,iPeri)
C
C     $: doda delec, add a single particle into the flow
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'
      TYPE(meshType)  :: mesh
      TYPE(CommonParticleDataType) :: cpd  ! common particle data
      TYPE(particleType) :: part(cpd%maxNp) ! seznam delcev
      TYPE(fluidType) :: fluid  ! hitrostno, vrtincno in T polje
      TYPE(fluidAtPartType) :: fap ! tekocina na mestu delca
      TYPE(penv) :: env
      REAL(8) nhh(mesh%nnodes,3)

      INTEGER i,id,ierr,cell,ppno,iPeri
      REAL(8) x,y,z,st,vx,vy,vz

c     na katerem procesorju
      CALL GetNextPartNo(ppno,env%nproc)
      part(i)%ppno=ppno

c     samo ce je moj
      IF (env%myproc.EQ.part(i)%ppno) THEN

        part(i)%id=id
        part(i)%x=x
        IF (cpd%simt.EQ.2) THEN  ! cellular flow
          part(i)%y=0.02D0 !y  ! SAMO ZA CELL FLOW
        ELSE
          part(i)%y=y
        END IF
        part(i)%z=z
        part(i)%cell=cell
        part(i)%st=st
        part(i)%iPeri=iPeri  ! zacne v prvi periodi

        CALL GetFFFap(mesh,fluid,nhh,fap,part,cpd,i,ierr)
        IF (ierr.EQ.0) THEN
c         set particle active, its velocity equal velocity in inp file
          part(i)%active=.TRUE.
          part(i)%still=.FALSE.
          part(i)%vx=vx
          part(i)%vy=vy
          part(i)%vz=vz
        ELSE
          part(i)%active=.FALSE.
        END IF
c       premer
        part(i)%dp=cpd%f_l*SQRT(18.0D0*cpd%f_rho/cpd%p_rho*part(i)%st/cpd%f_re)
c       settling velocity
        part(i)%vsx=part(i)%dp**2/18.0D0/cpd%f_visk/cpd%f_u0*cpd%g0*cpd%gx*(cpd%p_rho/cpd%f_rho-1.0D0)
        part(i)%vsy=part(i)%dp**2/18.0D0/cpd%f_visk/cpd%f_u0*cpd%g0*cpd%gy*(cpd%p_rho/cpd%f_rho-1.0D0)
        part(i)%vsz=part(i)%dp**2/18.0D0/cpd%f_visk/cpd%f_u0*cpd%g0*cpd%gz*(cpd%p_rho/cpd%f_rho-1.0D0)
c       particle Reynolds number
        part(i)%rep=part(i)%dp*cpd%f_u0/cpd%f_visk
      END IF

      END


C -----------------------------------------------------------------------------
      SUBROUTINE ppos(mesh,ic,xi,eta,zeta,x,y,z)
C
C     $: izracuna x,y,z iz stevilke celice in xi,eta,zeta
C        calculate x,y,z from xi,eta,zeta
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
      REAL(8) xi,eta,zeta,x,y,z
      INTEGER ic

      REAL(8) x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4
      REAL(8) x5,y5,z5,x6,y6,z6,x7,y7,z7,x8,y8,z8
          
      
          x1=mesh%x(mesh%idc(ic,6),1)
          y1=mesh%x(mesh%idc(ic,6),2)
          z1=mesh%x(mesh%idc(ic,6),3)          
          x2=mesh%x(mesh%idc(ic,7),1)
          y2=mesh%x(mesh%idc(ic,7),2)
          z2=mesh%x(mesh%idc(ic,7),3)          
          x3=mesh%x(mesh%idc(ic,2),1)
          y3=mesh%x(mesh%idc(ic,2),2)
          z3=mesh%x(mesh%idc(ic,2),3)          
          x4=mesh%x(mesh%idc(ic,3),1)
          y4=mesh%x(mesh%idc(ic,3),2)
          z4=mesh%x(mesh%idc(ic,3),3)          
          x5=mesh%x(mesh%idc(ic,5),1)
          y5=mesh%x(mesh%idc(ic,5),2)
          z5=mesh%x(mesh%idc(ic,5),3)          
          x6=mesh%x(mesh%idc(ic,8),1)
          y6=mesh%x(mesh%idc(ic,8),2)
          z6=mesh%x(mesh%idc(ic,8),3)          
          x7=mesh%x(mesh%idc(ic,1),1)
          y7=mesh%x(mesh%idc(ic,1),2)
          z7=mesh%x(mesh%idc(ic,1),3)          
          x8=mesh%x(mesh%idc(ic,4),1)
          y8=mesh%x(mesh%idc(ic,4),2)
          z8=mesh%x(mesh%idc(ic,4),3)  

      CALL lks2kks(xi,eta,zeta,x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4,
     &                         x5,y5,z5,x6,y6,z6,x7,y7,z7,x8,y8,z8,x,y,z)
          

      END
      
C -----------------------------------------------------------------------------
      SUBROUTINE rndIno(stevilo,mxst)
C
C     $: vrne nakljucno naravno stevilo med 1 in maxst
C        v zacetku programa daj CALL RANDOM_SEED
C
C -----------------------------------------------------------------------------
      REAL(8) x
      INTEGER stevilo,mxst
      
      CALL RANDOM_NUMBER(x)
      stevilo=INT(x*mxst)+1
      
      END
      
C -----------------------------------------------------------------------------
      SUBROUTINE rndXiEtaZeta(xi,eta,zeta)
C
C     $: vrne 3 nakljucna stevila med -1 in 1
C        v zacetku programa daj CALL RANDOM_SEED
C
C -----------------------------------------------------------------------------
      REAL(8) xi,eta,zeta,x
      
      CALL RANDOM_NUMBER(x)
      xi=-1.0D0+2.0D0*x

      CALL RANDOM_NUMBER(x)
      eta=-1.0D0+2.0D0*x
      
      CALL RANDOM_NUMBER(x)
      zeta=-1.0D0+2.0D0*x
            
     
      END     


C -----------------------------------------------------------------------------
      SUBROUTINE CoR_NeighbourList(env,mesh,io,inp)
C
C     $: Preberi ali izracunan na novo seznam sosedov
C        Read from file (.neil) or Calculate a list of neighbouring cells
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
      TYPE(IOtype)    :: io
      TYPE(inputtype) :: inp
      TYPE(penv) :: env
      
      INTEGER iok,dummy

c     maksimalno stevilo sosedov ene celice je 26      
c     v 27-ti stolpec shranim stevilo sostedov
      ALLOCATE (mesh%neil(mesh%nicell,27))

      CALL CheckNeilFile(io,mesh%nnodes,mesh%nbnodes,mesh%nicell,iok)

      IF (iok.EQ.1) THEN
c       Read neil from disk 
        IF (env%myproc.EQ.1) CALL WarnErr(io,inp,0,"CoR_NeighbourList","Reading neighbour list!",0)
        OPEN (io%neil,FILE=TRIM(io%neil_name),FORM='UNFORMATTED',STATUS='OLD')      
        READ(io%neil) dummy,dummy,dummy
        CALL rdIMat(mesh%neil,mesh%nicell,27,io%neil)
        CLOSE(io%neil)    
      ELSE
c       Calculate neil      
        IF (env%myproc.EQ.1) CALL WarnErr(io,inp,0,"CoR_NeighbourList","Making neighbour list!",0)
        CALL MakeNeighList(mesh)  
c       write neil to disk
        IF (env%myproc.EQ.1) THEN
          OPEN (io%neil,FILE=TRIM(io%neil_name),FORM='UNFORMATTED',STATUS='UNKNOWN')
          WRITE(io%neil) mesh%nnodes,mesh%nbnodes,mesh%nicell
          CALL WrIMat(mesh%neil,mesh%nicell,27,io%neil)
          CLOSE(io%neil)
        END IF
      END IF

      
      END

C -----------------------------------------------------------------------------
      SUBROUTINE CheckNeilFile(io,a,b,c,iok)
C
C     $: Preveri ce je Neil na disku za pravo mrezo
C
C ----------------------------------------------------------------------------- 
      INCLUDE 'inc_types.f'
      TYPE(IOtype) io 
      
      INTEGER a,b,c
      INTEGER aa,bb,cc,iok

      iok=0
      OPEN(io%neil,FILE=TRIM(io%neil_name),FORM='UNFORMATTED',STATUS='OLD',ERR=10)
      READ(io%neil) aa,bb,cc
            
      IF (a.EQ.aa.AND.b.EQ.bb.AND.c.EQ.cc) iok=1    
      CLOSE(io%neil)
      
10    RETURN
      END

C -----------------------------------------------------------------------------
      SUBROUTINE TestVelInter(mesh)
C
C     $: testira rutino ki naredi xi,eta,zeta  iz x,y,z
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
            
      REAL(8) xi,eta,zeta,x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4
      REAL(8) x5,y5,z5,x6,y6,z6,x7,y7,z7,x8,y8,z8,x,y,z
      INTEGER ierr            
      
      INTEGER ic

      ic=1
      
          x1=mesh%x(mesh%idc(ic,6),1)
          y1=mesh%x(mesh%idc(ic,6),2)
          z1=mesh%x(mesh%idc(ic,6),3)          
          x2=mesh%x(mesh%idc(ic,7),1)
          y2=mesh%x(mesh%idc(ic,7),2)
          z2=mesh%x(mesh%idc(ic,7),3)          
          x3=mesh%x(mesh%idc(ic,2),1)
          y3=mesh%x(mesh%idc(ic,2),2)
          z3=mesh%x(mesh%idc(ic,2),3)          
          x4=mesh%x(mesh%idc(ic,3),1)
          y4=mesh%x(mesh%idc(ic,3),2)
          z4=mesh%x(mesh%idc(ic,3),3)          
          x5=mesh%x(mesh%idc(ic,5),1)
          y5=mesh%x(mesh%idc(ic,5),2)
          z5=mesh%x(mesh%idc(ic,5),3)          
          x6=mesh%x(mesh%idc(ic,8),1)
          y6=mesh%x(mesh%idc(ic,8),2)
          z6=mesh%x(mesh%idc(ic,8),3)          
          x7=mesh%x(mesh%idc(ic,1),1)
          y7=mesh%x(mesh%idc(ic,1),2)
          z7=mesh%x(mesh%idc(ic,1),3)          
          x8=mesh%x(mesh%idc(ic,4),1)
          y8=mesh%x(mesh%idc(ic,4),2)
          z8=mesh%x(mesh%idc(ic,4),3)            
      
      x=1.0000D0
      y=0.0001D0
      z=0.5D0

      
      CALL kks2lks(x,y,z,x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4,
     &                   x5,y5,z5,x6,y6,z6,x7,y7,z7,x8,y8,z8,xi,eta,zeta,ierr)
     
      
      END


C -----------------------------------------------------------------------------
      SUBROUTINE InitLVC(env,mesh,io,part,cpd,inp,lvc)
C
C     $: init list of visited cells
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'


      TYPE(CommonParticleDataType) :: cpd  ! common particle data
      TYPE(particleType) :: part(cpd%maxNp) ! seznam delcev
      TYPE(meshType)  :: mesh
      TYPE(IOtype)    :: io
      TYPE(inputtype) :: inp
      TYPE(penv) :: env
      INTEGER lvc(mesh%nicell,inp%lvcs)

      INTEGER i,j,nd
      REAL(8) h

      lvc=0
      cpd%maxSt=-1.0D0
      cpd%minSt=+1.0D10
      DO i=1,cpd%cnp
        IF (part(i)%st.LT.cpd%minSt) cpd%minSt=part(i)%st
        IF (part(i)%st.GT.cpd%maxSt) cpd%maxSt=part(i)%st
      END DO
c       stokes razred delca, stevilo razredov = inp%lvcs
      h=(cpd%maxSt-cpd%minSt)/inp%lvcs
      DO i=1,cpd%cnp
        DO j=1,inp%lvcs
          IF (part(i)%st.GE.(cpd%minSt+h*(j-1)).AND.part(i)%st.LE.(cpd%minSt+h*j)) THEN
            part(i)%StClass=j
          END IF
        END DO
      END DO
c       naredimo statistiko stevila delcev po razredih
      IF (env%myproc.EQ.1) THEN
        WRITE (io%l,*) "StClass, MinSt, MaxSt, N particles"
        DO j=1,inp%lvcs
          nd=0
          DO i=1,cpd%cnp
            IF (part(i)%StClass.EQ.j) nd=nd+1
          END DO
          WRITE (io%l,*) j,cpd%minSt+h*(j-1),cpd%minSt+h*j,nd
        END DO
        WRITE (io%l,*) ""
      END IF

      END



C -----------------------------------------------------------------------------
      SUBROUTINE wPdens(mesh,pdens)
C
C     $: write particle density results file
C
C -----------------------------------------------------------------------------
      INCLUDE 'inc_types.f'

      TYPE(meshType)  :: mesh
      REAL(8) pdens(mesh%nnodes)

      CHARACTER*200 vrstica

      INTEGER itp,nic,i

      itp=77
      nic=8*mesh%nicell
      OPEN (itp,FILE="vreteno.pdens.dat",STATUS='UNKNOWN')

      WRITE (itp,'(A)') 'VARIABLES = "X", "Y", "Z", "node","pdens"'

      WRITE (itp,'(A,I8,A,I8)')
     &      'ZONE T="cell flow field",F=FEPOINT, ET=BRICK,N= ', mesh%NNODES, ',E=', nic

C       Output results
        DO  i=1,mesh%nnodes
          WRITE (vrstica,'(3G18.9,I8,G18.9)')
     &        mesh%x(i,1),mesh%x(i,2),mesh%x(i,3),i,
     &        pdens(i)

          CALL sqblnk(itp,vrstica)
        END DO
        CALL TECwIDC(itp,mesh)

      CLOSE(itp)

      END
