C
      PROGRAM VRETENO
C
C     Sledenje PartIcle TrackiNg DeLcEv
C     -        -   -          -  - - -
C
C     SPINDLE = VRETENO
C
C      
C     Written by
C
C           - Jure Ravnik, jure.ravnik@uni-mb.si
C
C           University of Maribor
C           Faculty of Mechanical Engineering
C           Smetanova 17, SI-2000, Maribor, Slovenia
C
C     Web page
C
C          http://iepoi.uni-mb.si/ravnik/vreteno.html
C
C
C     Description
C
C     Vreteno is a software simulation tool capable of simualting movement of
C     particles in fluid flows under the influence of hydrodynamic and magnetic forces
C
!
!     Licence
!
!     This software may be freely used provided that any results, that were obtained by the
!     use of this software or parts of it, cite the following papers:
!
!     RAVNIK, Jure, HRIBERŠEK, Matjaž. High gradient magnetic particle separation in viscous flows by 3D BEM.
!     Comput. mech., Online First, 24 May 2012, doi: 10.1007/s00466-012-0729-3.
!     http://dx.doi.org/10.1007/s00466-012-0729-3
!
!     RAVNIK, Jure, ŠKERGET, Leopold, HRIBERŠEK, Matjaž, ŽUNIČ, Zoran.
!     Numerical simulation of dilute particle laden flows by wavelet BEM-FEM.
!     Comput. methods appl. mech. eng.. [Print ed.], Jan. 2008, vol. 197, iss. 6/8, pp. 789-805.
!     http://dx.doi.org/10.1016/j.cma.2007.09.007.
!
!     where the theory behind this software is described.
!
!
!     Requirements
!
!
!     Makefile is written for Inter Fortran Compiler, but the code should work
!     with any FORTRAN 90 compiler. When linking, MPI library must be likned,
!     version 1.2.7 is used by the author. See Makefile in source folder for
!     further information.
!
!
!     Disclaimer of Warranty.
!
!     THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
!     APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
!     HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
!     OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
!     THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!     PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
!     IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
!     ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
!
!     Limitation of Liability.
!
!     IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
!     WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
!     THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
!     GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
!     USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
!     DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
!     PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
!     EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
!     SUCH DAMAGES.
!

C
C     The file inc_types.f holds all coustom type definitions, used by this program
C     This file must be included in almost all subroutines.
C
      INCLUDE 'inc_types.f'

C     mesh varible stores all information about the computational mesh
      TYPE(meshType)  :: mesh
C     io variable stores all information for handling input/output of the program
      TYPE(IOtype)    :: io
C     inp variable stores all information read from the main input file
      TYPE(inputtype) :: inp      
C     gauss variable stores gauss weights and points for numerical integration
      TYPE(gausstype) :: gauss
C     fstat,sumFstat store statistics about forces working on particles
      TYPE(ForceStatisticsType) :: fstat,sumFstat

C     cpd stores information common to all particles
      TYPE(CommonParticleDataType) :: cpd  ! common particle data
C     part is a list which stores location and velocity of each particle
      TYPE(particleType), ALLOCATABLE :: part(:) ! seznam delcev 
C     fluid stores fluid flow variables, velocity, vorticity, temperature fields on the computational grid
      TYPE(fluidType) :: fluid  ! hitrostno, vrtincno in T polje  
C     fap stores fluid flow variables at the location of the particle
      TYPE(fluidAtPartType) :: fap ! tekocina na mestu delca
C     env stores environmental data for parallel execution of the code
      TYPE(penv) :: env

c     list of visited cells
      INTEGER, ALLOCATABLE :: lvc(:,:),sumlvc(:,:)
c     list of cells, where magnetic force dominates
      INTEGER, ALLOCATABLE :: summfdom(:)
c     distance to closest wire (in case of magnetic field produced by thin wires)
      REAL(8), ALLOCATABLE :: mdist(:),sumMdist(:)
c     number of particles attached to wires (in case of magnetic field produced by thin wires in the flow)
      INTEGER, ALLOCATABLE :: nstill(:),sumNstill(:)

      
C     runge kutta variables
      REAL(8), ALLOCATABLE :: y(:),yout(:),dydx(:)
C
      REAL(8), ALLOCATABLE :: mfs(:,:) ! magnetic field strength H
      REAL(8), ALLOCATABLE :: nhh(:,:) ! \vec\nabla (\vec H\cdot\vec H)

      INTEGER ppno ! particle processor number
      LOGICAL done ! done indicator for each processor

      INTEGER n,icount,status
      INTEGER,ALLOCATABLE :: icontP(:)
      REAL(8) x,h ! time and timestep
      
      INTEGER i,j,ierr !,npart
      INTEGER st,sumst

c 
C____ Version data 
c
      CALL DatumInUra(inp%StartTime)
      inp%IDname='Vreteno'
      inp%IDversion='1.7.1'
      inp%IDdate='August 2012'
      CALL RANDOM_SEED
c
C____ MPI init (parallel initialization, needs MPI library instlled)
c
      CALL par_init(env)
c
C____ Set up Gauss quadrature positions and weights
c
      CALL GaussPosWeights(gauss)
c
C____ Input / output files
c
      CALL SetUpIO(env,io,inp)
      IF (env%myproc.EQ.1) THEN
        WRITE (io%l,*) "Number of processors",env%nproc ! io%l is the vreteno.log file
      END IF
c
C____ Read input file
c
      CALL Count_ReadInputFile(mesh,io,inp,cpd)
      IF (env%myproc.EQ.1) WRITE(io%l,*) "Number of PART statements",cpd%npartdesc
      ALLOCATE ( cpd%partdesc(cpd%npartdesc) )
      CALL ReadInputFile(mesh,io,inp,cpd)
c     Report parameters to log file
      IF (env%myproc.EQ.1) THEN
        WRITE (io%l,*) "Fluid Reynolds number     ",cpd%f_Re
        WRITE (io%l,*) "Fluid viscosity [m/s^2]   ",cpd%f_visk
        WRITE (io%l,*) "Fluid length scale [m]    ",cpd%f_L
        WRITE (io%l,*) "Fluid density [kg/m^3]    ",cpd%f_rho
        WRITE (io%l,*) "Fluid velocity scale [m/s]",cpd%f_u0
        IF (inp%imag.GT.0) WRITE (io%l,*) "Fluid magnetic susc.      ",cpd%f_hi
        WRITE (io%l,*) "Particle density [kg/m^3] ",cpd%p_rho
        IF (inp%imag.GT.0) WRITE (io%l,*) "Particle magnetic susc.   ",cpd%p_hi
        WRITE (io%l,*) "parameter R               ",cpd%R
        WRITE (io%l,*) "parameter A               ",cpd%A
        WRITE (io%l,*) "gravity X [m/s^2]         ",cpd%g0*cpd%gx
        WRITE (io%l,*) "gravity Y [m/s^2]         ",cpd%g0*cpd%gy
        WRITE (io%l,*) "gravity Z [m/s^2]         ",cpd%g0*cpd%gz
        IF (inp%imag.GT.0) WRITE (io%l,*) "Mag. field strength [A/m] ",cpd%H0
        IF (inp%imag.GT.0) WRITE (io%l,*) "Mag. field density  [T]   ",cpd%H0*1.2566368D-6
        WRITE (io%l,*) ""
      END IF
C
C     Read mesh
C
      CALL ReadMeshGid3D(io,inp,mesh,gauss)
C
C     Make list of cell neighbours (to speed up finding in which cell resides individual particle)
C
      CALL CoR_NeighbourList(env,mesh,io,inp)

C
C     Read or generate flow fields (based on SIMT defined in input file)
C
      IF (cpd%simt.EQ.1.or.cpd%simt.EQ.4.or.cpd%simt.EQ.5.or.cpd%simt.EQ.6) THEN
C       Read Tritok flow simulation result file (*.dat)
        CALL ReadResultsDataFile(mesh,io,inp,fluid)
      ELSE IF (cpd%simt.EQ.2) THEN  ! generate cellular flow
        CALL GenCellFlow(mesh,inp%xpi,fluid)
        IF (env%myproc.EQ.1) CALL wCellFlowTec(mesh,fluid)
      ELSE IF (cpd%simt.EQ.3) THEN  ! generate helical flow
        CALL GenHeliFlow(mesh,fluid,cpd)
        IF (env%myproc.EQ.1) CALL wHeliFlowTec(mesh,fluid)
      END IF

C
C     set up magnetic field based on MMODEL as set in the input file
C
      ALLOCATE (nhh(mesh%nnodes,3))
      IF (inp%imag.GT.0) THEN
        ALLOCATE (mfs(mesh%nnodes,3))
        CALL SetUpMagField(mesh,cpd,inp%mmodel,inp%xpi,mfs,nhh,inp)
        IF (env%myproc.EQ.1) THEN
          CALL wMagFieldTec(mesh,mfs,nhh)
          WRITE (io%l,*) "Magnetic factor   ",cpd%mfac
        END IF
      ELSE
        nhh=0.0D0
        cpd%mfac=0.0D0
      END IF
c
c     set up initial particle positions
c
      ALLOCATE (part(cpd%maxNp))
      cpd%cnp=0  ! current number of particles =0
      cpd%lid=0  ! last particle ID
      ppno=0 ! previous particle number
      CALL SetUpParticles(env,mesh,fluid,nhh,part,cpd,0,ppno)
C   
C     list of visited cells
C      
      IF (inp%lvcs.GT.0) THEN
        ALLOCATE (lvc(mesh%nicell,inp%lvcs))
        CALL InitLVC(env,mesh,io,part,cpd,inp,lvc)
      END IF
c
c     Force statistics
c
      ALLOCATE (fstat%gb(mesh%nicell))
      ALLOCATE (fstat%drag(mesh%nicell))
      ALLOCATE (fstat%adpc(mesh%nicell))
      ALLOCATE (fstat%mag(mesh%nicell))
      ALLOCATE (fstat%vabs(mesh%nicell))
      ALLOCATE (fstat%aabs(mesh%nicell))
      ALLOCATE (fstat%n(mesh%nicell))
      fstat%gb=0.0D0
      fstat%drag=0.0D0
      fstat%adpc=0.0D0
      fstat%mag=0.0D0
      fstat%vabs=0.0D0
      fstat%aabs=0.0D0
      fstat%n=0
c
c     monitor force for one particle per processor (MONF keyword)
c     vsak procesor sledi enemu delcu skozi cas
      IF (inp%mfpid.GT.0) THEN
c       open monitor forces output file
        OPEN (io%imf,FILE=io%imf_name,STATUS='UNKNOWN')
        WRITE (io%imf,'(A)') 'VARIABLES="time","gb","drag","adps","mag","drag/gb","adps/gb","mag/gb","uminvabs"'
c       finds particle to follow (monitor)
        DO i=1,cpd%cnp
          IF (env%myproc.EQ.part(i)%ppno) THEN ! vzame prvi njegov delec
            inp%mfpid=i
            EXIT
          END IF
        END DO
      END IF
c
c     Start time integration
c
c     time step
      IF (env%myproc.EQ.1) THEN
        WRITE (io%l,*) "time step            ",inp%tstep
        WRITE (io%l,*) "time step [s]        ",inp%tstep*cpd%f_L/cpd%f_u0
        WRITE (io%l,*) "number of time steps ",inp%nsteps
      END IF

      n=6 ! stevilo enacb, number of equations to integrate by Runge-Kutta
      ALLOCATE (y(n),yout(n),dydx(n))  ! Runge - Kutta variables
      x=0.0D00    ! cas, time
      h=inp%tstep    ! delta t, time step

c     open binary results files
      icount=0 ! koliko timestepov bom izpisal
      icount=icount+1 ! number of time steps in binary files

c     export tecplot particle location (main result file), (PLOC keyword)
      IF (inp%ploc.GT.0) THEN
        OPEN (io%ibin+env%myproc,FILE=TRIM(io%ibin_name),FORM='UNFORMATTED',STATUS='UNKNOWN')
        WRITE (io%ibin+env%myproc) env%nproc,env%myproc
c       write initial particle locations into results file
        CALL outBIN(env,x,io,cpd,part)
      END IF

c     monitor particle density (PDEN keyword)
      IF (inp%pdens.GT.0) THEN
        OPEN (io%ipdens,FILE=TRIM(io%ipdens_name),FORM='UNFORMATTED',STATUS='UNKNOWN')
        WRITE (io%ipdens) env%nproc,env%myproc
c       write initial particle density into results file
        CALL OutBINpdens(io,fluid%pdens,mesh%nnodes,x)
      END IF
c
c     set up minimal distance to wire, magnetic model 6 only
      IF (inp%mmodel.EQ.6) THEN
        ALLOCATE(mdist(inp%nsteps))
        mdist=100.1D0 ! prevelika, ce ostane taka to pomeni, da je zadel zico
      END IF

c     count number of still particles in each timestep
      ALLOCATE(nstill(inp%nsteps))
      nstill=0

c     I am not done yet
      done=.FALSE.

      j=0 ! count time steps
c
c     main time loop
c
      DO WHILE (j.LT.inp%nsteps.AND.done.EQ..FALSE.)
        j=j+1  ! current time step
        x=x+h  ! increase time
        fluid%pdens=0.0D0
c       loop over particles
        DO i=1,cpd%cnp
c         if the particle is active, not still and assigned to my processor then
          IF ((part(i)%active.EQ..TRUE.).AND.(part(i)%still.EQ..FALSE.).AND.(env%myproc.EQ.part(i)%ppno)) THEN
123         CONTINUE
c           get fluid properties at particle
            CALL GetFFFap(mesh,fluid,nhh,fap,part,cpd,i,ierr) ! to bi moral imeti dvakrat
            IF (ierr.NE.0) THEN ! particle is not located inside of the computational domain
              IF (cpd%simt.EQ.1) THEN
                part(i)%active=.FALSE.  ! since it is not it the domain, it is not active anymore
              ELSE IF (cpd%simt.EQ.6) THEN ! KANAL
                part(i)%active=.FALSE.
              ELSE IF (cpd%simt.EQ.2) THEN  ! cellular flow
                IF (part(i)%x.GT.1.0D0) part(i)%x=part(i)%x-1.0D0  ! here periodic BC are applied
                IF (part(i)%x.LT.0.0D0) part(i)%x=part(i)%x+1.0D0
                IF (part(i)%y.GT.1.0D0) part(i)%y=part(i)%y-1.0D0
                IF (part(i)%y.LT.0.0D0) part(i)%y=part(i)%y+1.0D0
                IF (part(i)%z.GT.1.0D0) part(i)%z=part(i)%z-1.0D0
                IF (part(i)%z.LT.0.0D0) part(i)%z=part(i)%z+1.0D0
c               ce ga magnet vrze samo toliko ven, da pride nazaj not na potem ostane not
c               ce ga vrze vec kot dve obmocji stran, potem ga izgubimo
                CALL GetFFFap(mesh,fluid,nhh,fap,part,cpd,i,ierr)
                IF (ierr.NE.0) THEN ! ce ga drugic ni, potem ga je magnet potisnil ven in je still
                  part(i)%active=.FALSE.
                  part(i)%still=.TRUE.
                END IF
              ELSE IF (cpd%simt.EQ.3) THEN  ! heli flow
                CALL HeliBC(part(i))
                GOTO 123                
              END IF
            END IF
          
            IF (inp%lvcs.GT.0) lvc(part(i)%cell,part(i)%StClass)=1
            IF (inp%imag.GT.0.AND.inp%mmodel.EQ.6.AND.part(i)%still.EQ..FALSE.) CALL ZadaneZico(part(i),inp%xpi)
            IF (inp%mmodel.EQ.6.AND.part(i)%still.EQ..FALSE.) THEN
              CALL RazdaljaDoNajblizjeZice(part(i),inp%xpi,mdist(j))
            END IF
          END IF   


          IF ((part(i)%active.EQ..TRUE.).AND.(part(i)%still.EQ..FALSE.).AND.(env%myproc.EQ.part(i)%ppno)) THEN
c           zacetni pogoji
            y(1)=part(i)%x  ! zacetni polozaj x, old position (x direction)
            y(2)=part(i)%vx ! zacetna hitrost x, old velocity (x direction)
            y(3)=part(i)%y ! zacetni polozaj y old position (y direction)
            y(4)=part(i)%vy ! zacetna hitrost y, old velocity (y direction)
            y(5)=part(i)%z ! zacetni polozaj z old position (z direction)
            y(6)=part(i)%vz ! zacetna hitrost z, old velocity (z direction)
            CALL cal_dvdt(i,part,cpd,fap,y,dydx)
c           spremljamo sile na delec s casom
c           monitor forces on partices (MONF keyword)
            IF (inp%mfpid.GT.0) CALL mon_forces(i,part,cpd,fap,y,x,io,mesh,inp%mfpid,fstat,dydx)
c           Runge Kutta
            CALL rk4(y,dydx,n,x,h,yout,i,part,cpd,fap)          
c           new location and veloctiy
            part(i)%x=yout(1)  
            part(i)%vx=yout(2) 
            part(i)%y=yout(3)
            part(i)%vy=yout(4)
            part(i)%z=yout(5)
            part(i)%vz=yout(6)  
          END IF

c         Apply particle boundary conditions for different simulation types (SIMT keyword)
          IF (cpd%simt.EQ.4) THEN  ! driven cavity
            CALL CavityBC(part(i)) ! ce je zgoraj, potem zapise delec zunaj
          END IF   
          IF (cpd%simt.EQ.5) THEN  ! driven cavity
            CALL CavitydpBC(cpd,part(i)) ! ce je zgoraj, potem zapise delec zunaj
          END IF   
          IF (cpd%simt.EQ.6) THEN  ! channel flow
            CALL ChannelBC(part(i),inp%nPeri,inp%Xlength)
          END IF

c         count still particles
          IF (part(i)%still.EQ..TRUE.) THEN
            nstill(j)=nstill(j)+1
          END IF
        END DO ! end main loop over particles

c       before starting a new timestep, set up new particles, based on PART statements in input file
        CALL SetUpParticles(env,mesh,fluid,nhh,part,cpd,j,ppno)


c       vsak procesor ima drugo stevilo timestepov (icount), to mora CopyBIN2TXT nekako upostevati
c       output timestep results into binary files
        IF (MOD(j,inp%outst).EQ.0.OR.j.EQ.inp%nsteps) THEN            ! izpis rezultatov
          icount=icount+1  ! icount = number of timesteps written in binary file
          IF (inp%ploc.GT.0) CALL outBIN(env,x,io,cpd,part)
          IF (inp%pdens.GT.0) CALL OutBINpdens(io,fluid%pdens,mesh%nnodes,x)
        END IF
        IF (MOD(j,inp%nsteps/20).EQ.0.AND.env%myproc.EQ.1) THEN  ! izpis v log file, na vsakih 5%
          WRITE (io%l,*) "time=",x
        END IF

c       check if I am done
        done=.TRUE.
        DO i=1,cpd%cnp
          IF ((part(i)%active.EQ..TRUE.).AND.(part(i)%still.EQ..FALSE.).AND.(env%myproc.EQ.part(i)%ppno)) THEN
            done=.FALSE.
            exit
          END IF
        END DO

      END DO ! end main time loop

c     output final in-depth info on last particle position
      IF (inp%ploc.GT.0) THEN
        CALL outBINfinal(env,x,io,cpd,part)
c       close binary output file
        CLOSE (io%ibin+env%myproc)
      END IF

c     close particle density output files
      IF (inp%pdens.GT.0) THEN
        CLOSE (io%ipdens)
      END IF

c     pocakamo, da vsi procesorji zapisejo svoje bin datoteke
c     wait until all processors write their binary files
      CALL MPI_BARRIER(env%comm,ierr)
c     copy binary output to text output
c     communicate icount (number of time steps in binary result file)
c     ne dela send / revc preveri
      IF (env%myproc.EQ.1) THEN
        ALLOCATE (icontP(env%nproc))
        icontP=0
        icontP(1)=icount
        DO i=2,env%nproc
          CALL MPI_RECV(icontP(i),1,MPI_INTEGER,i-1,1,env%comm,status, IERR)
        END DO
      ELSE
        CALL MPI_SEND(icount,1,MPI_INTEGER,0,1,env%comm, IERR)
      END IF

c     Processor number 1 writes ASCII result file out of individual binary files
      IF (inp%ploc.GT.0.AND.env%myproc.EQ.1) CALL CopyBIN2TXT(env,io,icontP,inp)
c     output particle density distribution
      IF (inp%pdens.GT.0.AND.env%myproc.EQ.1) CALL CopyPdensBIN2TXT(env,io,mesh,icount)
c     output list of visited cells
      IF (inp%lvcs.GT.0) THEN
        ALLOCATE (sumlvc(mesh%nicell,inp%lvcs))
        icount=mesh%nicell*inp%lvcs
        CALL MPI_REDUCE(lvc, sumlvc, icount, MPI_INTEGER, MPI_SUM, 0, env%comm, IERR)
        IF (env%myproc.EQ.1) CALL Outlvc(mesh,io,inp,sumlvc)
      END IF
c
c     Postprocessing for magnetic model 6
c
      IF (inp%mmodel.EQ.6) THEN
c       count still
        st=0
        sumst=0
        DO i=1,cpd%cnp
          IF (env%myproc.EQ.part(i)%ppno) THEN
            IF (part(i)%still.EQ..TRUE.) st=st+1
          END IF
        END DO
        CALL MPI_REDUCE(st, sumst, 1, MPI_INTEGER, MPI_SUM, 0, env%comm, IERR)
        IF (env%myproc.EQ.1) WRITE (io%l,*) "still particles    ",sumst
c
c       Export mfdom
c
        ALLOCATE (sumMfdom(mesh%nicell))
        icount=mesh%nicell
        CALL MPI_REDUCE(mesh%mfdom, summfdom, icount, MPI_INTEGER, MPI_SUM, 0, env%comm, IERR)
        IF (env%myproc.EQ.1) CALL OutMfdom(mesh,io,sumMfdom)
        DEALLOCATE (sumMfdom)
      END IF
c
c     Force statistics
c
      IF (inp%mfpid.GT.0) THEN
        ALLOCATE (sumfstat%gb(mesh%nicell))
        ALLOCATE (sumfstat%drag(mesh%nicell))
        ALLOCATE (sumfstat%adpc(mesh%nicell))
        ALLOCATE (sumfstat%mag(mesh%nicell))
        ALLOCATE (sumfstat%vabs(mesh%nicell))
        ALLOCATE (sumfstat%aabs(mesh%nicell))
        ALLOCATE (sumfstat%n(mesh%nicell))
        icount=mesh%nicell
        CALL MPI_REDUCE(fstat%gb, sumfstat%gb, icount, MPI_DOUBLE_PRECISION, MPI_SUM, 0, env%comm, IERR)
        CALL MPI_REDUCE(fstat%drag, sumfstat%drag, icount, MPI_DOUBLE_PRECISION, MPI_SUM, 0, env%comm, IERR)
        CALL MPI_REDUCE(fstat%adpc, sumfstat%adpc, icount, MPI_DOUBLE_PRECISION, MPI_SUM, 0, env%comm, IERR)
        CALL MPI_REDUCE(fstat%mag, sumfstat%mag, icount, MPI_DOUBLE_PRECISION, MPI_SUM, 0, env%comm, IERR)
        CALL MPI_REDUCE(fstat%vabs, sumfstat%vabs, icount, MPI_DOUBLE_PRECISION, MPI_SUM, 0, env%comm, IERR)
        CALL MPI_REDUCE(fstat%aabs, sumfstat%aabs, icount, MPI_DOUBLE_PRECISION, MPI_SUM, 0, env%comm, IERR)
        CALL MPI_REDUCE(fstat%n, sumfstat%n, icount, MPI_INTEGER, MPI_SUM, 0, env%comm, IERR)
        DO i=1,mesh%nicell
          IF (sumfstat%n(i).GT.0) THEN  ! izracunam povprecja
            sumfstat%gb(i)=sumfstat%gb(i)/sumfstat%n(i)
            sumfstat%drag(i)=sumfstat%drag(i)/sumfstat%n(i)
            sumfstat%adpc(i)=sumfstat%adpc(i)/sumfstat%n(i)
            sumfstat%mag(i)=sumfstat%mag(i)/sumfstat%n(i)
            sumfstat%vabs(i)=sumfstat%vabs(i)/sumfstat%n(i)
            sumfstat%aabs(i)=sumfstat%aabs(i)/sumfstat%n(i)
          END IF
        END DO
        IF (env%myproc.EQ.1) CALL OutFstat(mesh,io,sumFstat)
      END IF
c
c     Minimal distance to wire
c
      IF (inp%mmodel.EQ.6) THEN
c       tu ima vsak procesor svoje razdalje, tiste ki ostanejo nad 100 so od drugih
c       ce jih dam na nic, lahko potem samo sestevam
        ALLOCATE (summdist(inp%nsteps))
        CALL MPI_REDUCE(mdist, summdist, inp%nsteps, MPI_DOUBLE_PRECISION, MPI_MIN, 0, env%comm, IERR)
        DO i=1,inp%nsteps
          IF (summdist(i).GT.100.0D0) summdist(i)=0.0D0
        END DO
        IF (env%myproc.EQ.1) CALL OutMdist(io,inp,sumMdist)
      END IF
c
c     Output number of still particles
c
      IF (inp%imag.GT.0.AND.inp%mmodel.NE.8.AND.inp%mmodel.NE.9.AND.inp%mmodel.NE.10) THEN
        ALLOCATE (sumnstill(inp%nsteps))
        icount=inp%nsteps
        CALL MPI_REDUCE(nstill, sumnstill, icount, MPI_INTEGER, MPI_SUM, 0, env%comm, IERR)
        IF (env%myproc.EQ.1) CALL OutNstill(io,inp,sumNstill,cpd)
      END IF
c
C____ Finish
c      
      CALL StopProgram(env,io,inp,0)
      END

  
